#!/usr/bin/env python3
# This file is based on TAoFI-Analyzer which is released under the GPL3 license.
# Go to https://github.com/raelize/TAoFI-Analyzer/LICENSE for full license details.

import argparse
from argparse import RawTextHelpFormatter
import plotly.express as px
import pandas as pd
import sqlite3
import re
import sys
from operator import itemgetter
from contextlib import closing
import os
import json
import base64

from dataclasses import dataclass, asdict, field
from dataclasses_json import dataclass_json
from typing import Dict

from os import listdir
from dash import Dash, dcc, html, Input, Output, State
from dash.exceptions import PreventUpdate
import dash_bootstrap_components as dbc
from dash import callback_context as ctx

from dash_ag_grid import AgGrid

_RECORDS = None
_COLORS = ['green', 'yellow', 'magenta', 'orange', 'cyan', 'blue', 'black', 'red']

_COLOR_CONFIG = {
    'P': ('pink', 'black', 'timeout'),
    'G': ('green', 'white', 'green'),
    'Y': ('yellow', 'black', 'yellow'),
    'M': ('magenta', 'white', 'magenta'),
    'O': ('orange', 'white', 'orange'),
    'C': ('cyan', 'white', 'cyan'),
    'B': ('blue', 'white', 'blue'),
    'Z': ('black', 'white', 'black'),
    'R': ('red', 'white', 'red')
}

@dataclass_json
@dataclass
class AnalyzerConfig:
    serverip: str = "127.0.0.1"
    serverport: int = 8080
    directory: str = None
    database: str = None
    y: str = None
    x: str = None
    argv: str = None
    database: str = None
    query: str = ''
    colors: Dict[str, str] = field(default_factory=dict)

def update_legend_labels(fig,labels):
    for entry in fig.data:
        if entry['name'] in labels:
            entry['name'] = labels[entry['name']]

def get_number_of_experiments(directory, database):
    database_path = os.path.join(directory, database)

    try:
        with closing(sqlite3.connect(database_path)) as connection:
            with closing(connection.cursor()) as cursor:
                cursor.execute("SELECT COUNT(*) FROM experiments")
                return cursor.fetchone()[0]
    except Exception as e:
        print("ERROR (get_number_of_experiments): %s" %(e))

# TODO: add date
def get_databases(directory):
    # get all databases in directory
    databases = []
    for file in listdir(directory):
        if re.search('^.*\\.sqlite$',file):
            databases.append(file)
    databases.sort(reverse=True)

    # transform to options
    databases_options = []
    for index in range(len(databases)):
        label = "%s (%d)" %(databases[index], get_number_of_experiments(directory, databases[index]))
        databases_options.append( {'label':label, 'value': databases[index]} )

    return databases_options

def get_argv(directory, database):
    database_path = os.path.join(directory, database)

    try:
        with closing(sqlite3.connect(database_path)) as connection:
            with closing(connection.cursor()) as cursor:
                cursor.execute("SELECT argv FROM metadata")
                argvstr = cursor.fetchone()[0]
                return argvstr
    except Exception as e:
        print("ERROR (get_argv): %s" %(e))

def get_parameters(directory, database):
    database_path = os.path.join(directory, database)
    try:
        with closing(sqlite3.connect(database_path)) as connection:
            with closing(connection.cursor()) as cursor:
                cursor.execute("SELECT * FROM experiments")
                parameters = list(next(zip(*cursor.description)))
                parameters.remove('response')
                return parameters
    except Exception as e:
        print("ERROR (get_parameters): %s" %(e))

def run_app(config):
    app = Dash(__name__, external_stylesheets=[dbc.themes.LUX])
    app.css.config.serve_locally = True
    app.scripts.config.serve_locally = True

    app.layout = html.Div([
            dcc.Store(id='config-store', data=asdict(config)),
            html.Div([
                html.H4('Fault Injection Analysis'),
            ],style={'width':'80%','border-style':'none','margin':'0 auto'}),               
            html.Div([
                dbc.Card(
                    dbc.CardBody([

                        html.Div([
                            html.Button("Update", id='update-button', n_clicks=0, style={'width':'100px'}),
                            html.Datalist(id="examples", children=[
                                html.Option(value="match_string(response, 'ets')"),
                                html.Option(value="match_hex(response, '661b')"),
                                html.Option(value="color = 'G'"),
                                html.Option(value="delay > 100"),
                                html.Option(value="length > 100"),
                            ]),
                            dcc.Input(id='query-input', type="text", list='examples', value='', style={'width':'100%','display': 'inline-block'}, placeholder="SELECT * FROM experiments WHERE"),

                            dcc.Upload(
                                id="upload_data",
                                children=html.Button("Import", style={'width':'100px'}),
                                multiple=False,
                            ),

                            html.Button("Export", id='config-export-button',  style={'width':'100px', 'display': 'inline-block'}),
                            dcc.Download(id="download_data")
                        ], style={'display': 'flex', 'alignItems': 'center'})
                    ])
                ),

                dbc.Card(
                    dbc.CardBody([
                        dcc.Dropdown(id='database-dropdown', style={'width':'100%'}, options=get_databases(config.directory)),
                        dcc.Dropdown(id='x-dropdown', style={'width':'100%'}),
                        dcc.Dropdown(id='y-dropdown', style={'width':'100%'})
                    ])
                ),
                dbc.Card(
                    dbc.CardBody([
                        html.Center([
                            dcc.Graph(id='graph', style={'width':'80%'}),
                        ]),
                    ])
                ),
                dbc.Card(
                    dbc.CardBody([

                        html.P('re.search(*, response)'),

                            *(
                                input_component
                                for color in _COLORS
                                for input_component in [
                                    dcc.Input(
                                        id=f'recolor-{color}',type="text",placeholder=f"{color}", style={'width':'15%'}
                                    ),
                                    dcc.Input(
                                        id=f'recolor-{color}-label',type="text",placeholder=f"{color}-label",style={'width':'15%', 'margin-right': '10px', 'margin-bottom':'10px'}
                                    )
                                ]
                            ),
                    ])
                ),
                dbc.Card(
                    dbc.CardBody([
                        dbc.Switch(
                            id='switch-squeezedata',
                            value=False,
                            label='Squeeze Data',
                            style={'display': 'inline-block', 'marginRight': '20px'}
                        ),
                        dbc.Switch(
                            id='switch-showhexdata',
                            value=False,
                            label='Show Hex',
                            style={'display': 'inline-block'}
                        ),
                        html.Div(id='data',style={'width':'100%', 'height':'100%', 'border-style':'none'}),
                    ])
                ),
                dbc.Card([
                    dbc.CardHeader('Arguments:'),
                    dbc.CardBody([
                        dcc.Markdown('', id='argv'),
                    ])
                ]),
                dbc.Card([
                    dbc.CardHeader('Points:'),
                    dbc.CardBody([
                        dcc.Markdown('', id='points'),
                    ])
                ]),
                dbc.Card([
                    dbc.CardHeader('Store:'),
                    dbc.CardBody([
                        dcc.Markdown('', id='printstore'),
                    ])
                ]),
            ],style={'width':'80%','border-style':'none','margin':'0 auto'}),
        ],style={'width':'100%', 'border-style':'none', 'margin-top':'100px','margin-bottom':'100px'})

    # callback for zoomed doints
    @app.callback(
        Output('points', 'children'),
        [
            Input('graph', 'relayoutData'),
            Input('graph', 'figure')
        ],
        prevent_initial_call=True
    )
    def zoomed_points(relayoutData, figure):
        if not figure or 'xaxis.range[0]' not in relayoutData:
            raise PreventUpdate

        layout = figure["layout"]
        x_axis = layout["xaxis"]
        y_axis = layout["yaxis"]

        if 'xaxis.range[0]' in relayoutData:
            ranges = {
                'x': (relayoutData['xaxis.range[0]'], relayoutData['xaxis.range[1]']),
                'y': (relayoutData['yaxis.range[0]'], relayoutData['yaxis.range[1]'])
            }
        else:
            ranges = {
                'x': tuple(x_axis['range']),
                'y': tuple(y_axis['range'])
            }

        p = f'''
            * {x_axis['title']['text']}
                * {ranges['x'][0]}
                * {ranges['x'][1]}
            * {y_axis['title']['text']}
                * {ranges['y'][0]}
             * {ranges['y'][1]}
        '''

        return p

    # callback for exporting config
    @app.callback(
       Output('download_data', 'data'),
       Input('config-export-button', 'n_clicks'),
       State('config-store', 'data'),
       prevent_initial_call=True
    )
    def download(n_clicks,store):
       config = AnalyzerConfig(**store)
       return dict(content=config.to_json(), filename="config.json")

    # callback for printing store at the bottom
    @app.callback(
        Output('printstore', 'children'),
        Input('config-store', 'data')
    )
    def printstore(store):
        p = ""
        for key,value in store.items():
            p += f"* {key}:{value}\n"
        return p

    # callback for updating store
    @app.callback(
        Output('config-store', 'data'),
        Output("database-dropdown", "value"),
        Output("x-dropdown", "value"),
        Output("y-dropdown", "value"),
        [
            Input('update-button', 'n_clicks'),
            Input('upload_data', 'contents'),
            Input('query-input', 'value'),
            Input('database-dropdown', 'value'),
            Input('x-dropdown', 'value'),
            Input('y-dropdown', 'value')
        ],
        [
            State('config-store', 'data'),
        ]
    )
    def update_store(nr_of_clicks, contents, query, database, x, y, store):
        if ctx.triggered_id == 'upload_data':
            if contents:
                content_type, content_string = contents.split(',')
                config_dict = json.loads(base64.b64decode(content_string))
                config = AnalyzerConfig(**config_dict)
                update_global_records(config)
                return config_dict,config.database,config.x,config.y

        if database is None:
            raise PreventUpdate

        database = database.split(' ')[0]
        config = AnalyzerConfig(**store)
        config.database = database
        config.x = x
        config.y = y
        config.query = query
        config.argv = get_argv(config.directory, config.database)
        return asdict(config),config.database,config.x,config.y

    # callback for printing the argv string at the bottom
    @app.callback(
        Output('argv', 'children'),
        Input('config-store', 'data'),
        prevent_initial_call=True
    )
    def update_argv(store):
        config = AnalyzerConfig(**store)
        return config.argv

    # callback for x list
    @app.callback(
        Output('x-dropdown', 'options'),
        Input('database-dropdown', 'value'),
        State('config-store', 'data'),
        prevent_initial_call=True
    )
    def update_dropdown_x(database, store):
        config = AnalyzerConfig(**store)
        return get_parameters(config.directory, database)

    # callback for y list
    @app.callback(
        Output('y-dropdown', 'options'),
        Input('database-dropdown', 'value'),
        State('config-store', 'data'),
        prevent_initial_call=True
    )
    def update_dropdown_y(database, store):
        config = AnalyzerConfig(**store)
        return get_parameters(config.directory, database)

    # new function for sqlite3 query
    def match_string(response, token):
        if token.encode(errors='strict') in response:
            return True
        else:
            return False

    # new function for sqlite3 query
    def match_hex(response, token):
        if bytes.fromhex(token) in response:
            return True
        else:
            return False

    def recolor(record, regex, new_color):
        if regex in [None, '']:
            return record['color']
        elif re.search(regex.encode(), record['response']):
            return new_color
        else:
            return record['color']

    def glitch_parameter_present(record, parameter):
        if parameter in record and record[parameter] not in [0, None]:
            return True
        else:
            return False

    def generate_data(records, squeeze_records=False):
        has_length = glitch_parameter_present(records[0], 'length')
        has_power = glitch_parameter_present(records[0], 'power')

        new_records = []

        if not squeeze_records:

            for record in records:
                new_record = {}
                new_record['id'] = record['id']
                new_record['color'] = record['color']
                new_record['delay'] = record['delay']
                if has_length:
                    new_record['length'] = record['length']
                if has_power:
                    new_record['power'] = record['power']
                new_record['rlen'] = len(record['response'])
                new_record['response'] = record['response'].decode('utf-8', errors='replace')
                new_record['hex(response)'] = record['response'].hex(' ')
                new_records.append(new_record)

            return new_records
        else:

            squeezed_records = {}

            for record in records:
                response = record['response'].decode('utf-8', errors='replace')

                if response not in squeezed_records:
                    squeezed_records[response] = {}
                    squeezed_records[response]['amount'] = 1
                    squeezed_records[response]['color'] = record['color']
                    squeezed_records[response]['Min(Delay)'] = record['delay']
                    squeezed_records[response]['Max(Delay)'] = record['delay']
                    if has_length:
                        squeezed_records[response]['Min(Length)'] = record['length']
                        squeezed_records[response]['Max(Length)'] = record['length']
                    if has_power:
                        squeezed_records[response]['Min(Power)'] = record['power']
                        squeezed_records[response]['Max(Power)'] = record['power']
                    squeezed_records[response]['response'] = response
                    squeezed_records[response]['hex(response)'] = record['response'].hex(' ')
                else:
                    squeezed_records[response]['amount'] += 1
                    squeezed_records[response]['Min(Delay)'] = min(squeezed_records[response]['Min(Delay)'], record['delay'])
                    squeezed_records[response]['Max(Delay)'] = max(squeezed_records[response]['Max(Delay)'], record['delay'])
                    if has_length:
                        squeezed_records[response]['Min(Length)'] = min(squeezed_records[response]['Min(Length)'], record['length'])
                        squeezed_records[response]['Max(Length)'] = max(squeezed_records[response]['Max(Length)'], record['length'])
                    if has_power:
                        squeezed_records[response]['Min(Power)'] = min(squeezed_records[response]['Min(Power)'], record['power'])
                        squeezed_records[response]['Max(Power)'] = max(squeezed_records[response]['Max(Power)'], record['power'])

            return sorted(squeezed_records.values(), key=itemgetter('amount'), reverse=True)

    def give_xy_label(parameter):
        labels = { 'length': '(ns)', 'delay': '(ns)', 'power': '(%)' }
        return labels.get(parameter, '')

    def update_global_records(config):
        global _RECORDS

        if not os.path.isfile(f"{config.directory}/{config.database}"):
            raise PreventUpdate

        con = sqlite3.connect(f"{config.directory}/{config.database}")

        # add some functions to sqlite
        con.create_function('match_string', 2, match_string)
        con.create_function('match_hex', 2, match_hex)

        # perform the query based on the query extension
        if config.query == '':
            query = 'SELECT * FROM experiments;'
        else:
            query = f'SELECT * FROM experiments WHERE {config.query};'

        # read stuff from database
        try:
            df = pd.read_sql(query, con)
            con.close()
        except:
            raise PreventUpdate

        # store records from global
        _RECORDS = df.to_dict('records')

    # callback graph; chained from update_store()
    @app.callback(
        Output('graph','figure'),
        [Input('config-store', 'data'), Input('x-dropdown', 'value'), Input('y-dropdown', 'value')],
        [State(f'recolor-{color}', 'value') for color in _COLORS] +
        [State(f'recolor-{color}-label', 'value') for color in _COLORS],
        prevent_initial_call=True
    )
    def update_graph(store, x, y, *color_states):
        global _RECORDS

        config = AnalyzerConfig(**store)

        if ctx.triggered_id == 'config-store':
            x = config.x
            y = config.y

        # prevent update
        if any(v is None for v in [x, y]):
            raise PreventUpdate

        update_global_records(config)

        # color amounts
        colors = { 'P':0,'G':0,'Y':0,'M':0,'O':0,'C':0,'B':0,'Z':0,'R':0 }

        color_values = color_states[:8]
        color_labels = color_states[8:]

        color_map = dict(zip(_COLORS,['G', 'Y', 'M', 'O', 'C', 'B', 'Z', 'R']))

        # recolor if needed
        for record in _RECORDS:
           for value, color_code in zip(color_values, color_map.values()):
               record['color'] = recolor(record, value, color_code)
           colors[record['color']] += 1

        # output plot
        try:
            fig = px.scatter(
                _RECORDS,
                x = x,
                y = y,
                render_mode = "webgl",
                color = "color",
                labels = {
                    'color': f'Classification ({len(_RECORDS):,})',
                    x: f'{x} {give_xy_label(x)}',
                    y: f'{y} {give_xy_label(y)}'
                },
                color_discrete_map = {
                   "P": "pink", "G": "green", "Y": "yellow", "M": "magenta",
                   "O": "orange", "C": "cyan", "B": "blue", "Z": "black", "R": "red"
                },
                category_orders = {"color" : ["P", "G","Y","M","O","C","B","Z","R"]}
            )
        except:
            raise PreventUpdate

        # update title of graph
        fig.update_layout(title_text=config.database[:-7], title_x=0.5, title_y=0.95)

        if config.x == 'x' or config.y == 'y':
            fig.update_xaxes(title_standoff=0, side='top')
            fig.update_yaxes(title_standoff=0, autorange='reversed')

        # Update legend labels
        labels = {}
        for color_code, label in zip(color_map.values(), color_labels):
           count = colors[color_code]
           labels[color_code] = f'{label} ( {count} / {count/len(_RECORDS):.1%} )'
        labels['P'] = f'timeout ( {colors["P"]} / {colors["P"]/len(_RECORDS):.1%} )'
        update_legend_labels(fig, labels)

        return fig

    # callback data; chained from update_graph()
    @app.callback(
        Output('data', 'children'),
        [
            Input('config-store', 'data'),
            Input('graph', 'figure'),
            Input('switch-squeezedata', 'value'),
            Input('switch-showhexdata', 'value')
        ],
        prevent_initial_call=True
    )
    def update_data(store, figure, squeeze, showhex):
        global _RECORDS

        if any(x is None for x in [figure, _RECORDS]):
            raise PreventUpdate

        # squeeze data (or not)
        data = generate_data(_RECORDS, squeeze_records=squeeze)

        # get columns from _RECORDS
        columns = data[0].keys()

        fields = []
        configs = []

        for column in columns:
            fields.append(column)
            if column in ['id', 'color', 'delay', 'length', 'power', 'rlen']:
                configs.append({
                    'autoSize':True,
                    'maxWidth': 100,
                    'cellStyle': {'textAlign': 'center'},
                   'headerClass': 'header-center-aligned',
                   'pinned': 'left'
                })
            elif 'Max' in column or 'Min' in column or column == 'amount':
                configs.append({
                    'autoSize':True,
                    'maxWidth': 150,
                    'cellStyle': {'textAlign': 'center'},
                    'headerClass': 'header-center-aligned',
                    'pinned': 'left'
                })
            elif column in ['response']:
                configs.append({
                    'autoSize':False,
                    'width': 500,
                    'flex': 1
                })
            elif column in ['hex(response)']:
                configs.append({
                    'autoSize':False,
                    'width': 500,
                    'hide': not showhex
                })
            else:
                configs.append({'autoSize':True,})

        columnDefs = [
            {'field': field, **config}
            for field, config in zip(fields, configs)
        ]

        # getRowStyle = {
        #     "styleConditions": [
        #         {
        #             "condition": "params.data.color = 'G'",
        #             "style": {"backgroundColor": "green"},
        #         },
        #     ],
        #     "defaultStyle": {"backgroundColor": "grey", "color": "white"},
        # }

        rowstyles = {
            "styleConditions": [
                {
                    "condition": "params.data.color == 'G'",
                    "style": {"backgroundColor": "#d5f5e3"},
                },
                {
                    "condition": "params.data.color == 'R'",
                    "style": {"backgroundColor": "#fadbd8"},
                },
                {
                    "condition": "params.data.color == 'Y'",
                    "style": {"backgroundColor": "#fcf3cf"},
                },
                {
                    "condition": "params.data.color == 'B'",
                    "style": {"backgroundColor": "#d4e6f1"},
                },
                {
                    "condition": "params.data.color == 'M'",
                    "style": {"backgroundColor": "#ebdef0"},
                },
                {
                    "condition": "params.data.color == 'O'",
                    "style": {"backgroundColor": "#fae5d3"},
                },
                {
                    "condition": "params.data.color == 'Z'",
                    "style": {"backgroundColor": "#d6dbdf"},
                },
            ],
            "defaultStyle": {"backgroundColor": "white", "color": "black"}
        }

        data = AgGrid(
            columnDefs=columnDefs,
            rowData=data,
            defaultColDef={
                'resizable': True,
                'sortable': True,
                'filter': True,
                'checkboxSelection': False
            },
            className='ag-theme-quartz',
            getRowStyle=rowstyles,
            dashGridOptions= {
                # 'groupHeaderHeight': 75,
                # 'headerHeight': 150,
                # 'floatingFiltersHeight': 40,
                'pagination': True,
                'animateRows': False
            },
            style={'height': '1000px'},
        )

        return data

    app.run(host=config.serverip, port=config.serverport, debug=True)

def run(args):
    config = AnalyzerConfig(
        serverip=args.ip,
        serverport=args.port,
        directory=args.directory,
        x = args.x,
        y = args.y,
    )

    for color in _COLORS:
        config.colors[color] = [None,None]

    run_app(config)

def main(argv=sys.argv):
    __version__ = "2.0"

    parser = argparse.ArgumentParser(
        description="analyzer.py v%s - Fault Injection Analyzer\nThis program is based on https://github.com/raelize/TAoFI-Analyzer." % __version__,
        prog="analyzer",
        formatter_class=RawTextHelpFormatter
    ) 
    parser.add_argument("--ip",help="Server port", type=str, default="127.0.0.1")
    parser.add_argument("--port",help="Server port", type=int, default=8080)
    parser.add_argument("--directory",help="Database directorys", required=True)
    parser.add_argument("--x", required=False, help="Preset the x parameter")
    parser.add_argument("--y", required=False, help="Preset the y parameter")

    args = parser.parse_args()

    run(args)

if __name__ == "__main__":
    main()


#!/usr/bin/env python3
# Copyright (C) 2024 Dr. Matthias Kesenheimer - All Rights Reserved.
# This file is based on TAoFI-FaultLib which is released under the GPL3 license.
# Go to https://github.com/raelize/TAoFI-FaultLib/LICENSE for full license details.
#
# You should have received a copy of the GPL3 license with this file.
# If not, please write to: m.kesenheimer@gmx.net.

"""
GeneticAlgorithm

This module includes all classes and functions that are necessary to search for optimal parameters in a multidimenstional parameter space.
"""

import random

class Parameterspace():
    def __init__(self, parameter_boundaries:list[tuple[float, float]], parameter_divisions:list[int]):
        # sanity checks
        if len(parameter_boundaries) != len(parameter_divisions):
            raise Exception(f"Error: parameter_boundaries and parameter_divisions have different lengths ({len(parameter_boundaries)} vs. {len(parameter_divisions)}).")
        for tup in parameter_boundaries:
            if tup[0] > tup[1]:
                raise Exception(f"Error: Upper bound {tup[1]} is lower than lower bound {tup[0]}.")
        self.parameter_boundaries = parameter_boundaries
        self.parameter_divisions = parameter_divisions
        self.cardinality = 1
        for num in self.parameter_divisions:
            self.cardinality *= num
        self.weights_per_bin = [0 for x in range(self.cardinality)]

    def get_cardinality(self):
        return self.cardinality

    def get_bin_assignment(self, *parameter:float) -> list[int]:
        fact = 1
        bina = 0
        for i in range(len(self.parameter_divisions)):
            division =  self.parameter_divisions[i]
            # xdelta = (xmax - xmin) / xdiv
            delta = (self.parameter_boundaries[i][1] - self.parameter_boundaries[i][0]) / division
            bin_number = int((parameter[i] - self.parameter_boundaries[i][0]) / delta)
            #print(f"par = {parameter[i]}, delta = {delta}, bina = {bin_number}")
            if i > 0:
                fact *= self.parameter_divisions[i - 1]
            bina += fact * bin_number
        return bina

    def add_experiment(self, weight:int, *parameter:float):
        for i in range(len(self.parameter_divisions)):
            if parameter[i] < self.parameter_boundaries[i][0] or parameter[i] >= self.parameter_boundaries[i][1]:
                print("[-] Error: parameter out of bounds. Skipping.")
                print(f"[-] Dimension: {i}, lower = {self.parameter_boundaries[i][0]}, upper = {self.parameter_boundaries[i][1]}, parameter = {parameter[i]}")
                return
        bina = self.get_bin_assignment(*parameter)
        self.weights_per_bin[bina] += weight

    def get_weights(self) -> list[int]:
        return self.weights_per_bin

    def get_bin_numbers_sorted_by_weights(self) -> list[int]:
        return sorted(range(len(self.weights_per_bin)), key=lambda i: self.weights_per_bin[i])

    def get_coordinates(self, bin_assignment:int) -> list[int]:
        bin_numbers = []
        reversed_parameter_divisions = list(reversed(self.parameter_divisions))
        fact = self.cardinality
        if bin_assignment > fact:
            raise Exception("Error: bin number exceeds total number of bins.")
        for div in reversed_parameter_divisions:
            c = (fact / div)
            b = int(bin_assignment / c)
            #print(f"bin_assignment = {bin_assignment}, fact = {fact}, div = {div}, (fact / div) = {c}, bin_number = {b}")
            bin_numbers.append(b)
            fact /= div
            bin_assignment -= int(b * c)
        return list(reversed(bin_numbers))

    def get_boundaries_from_coordinates(self, coordinates:list[float]) -> list[tuple[float, float]]:
        boundaries = []
        for i in range(len(self.parameter_divisions)):
            division =  self.parameter_divisions[i]
            delta = (self.parameter_boundaries[i][1] - self.parameter_boundaries[i][0]) / division
            lower = self.parameter_boundaries[i][0] + delta * coordinates[i]
            upper = self.parameter_boundaries[i][0] + delta * (coordinates[i] + 1)
            tup = (lower, upper)
            boundaries.append(tup)
        return boundaries

    def get_boundaries(self, bin_assignment:int) -> list[tuple[float, float]]:
        bin_numbers = self.get_coordinates(bin_assignment)
        return self.get_boundaries_from_coordinates(bin_numbers)

class Individual():
    def __init__(self, parameters:list[int], max_age:int = 10):
        self.parameters = parameters
        self.health = 0
        self.max_age = max_age
        self.age = 0

    def set_genom(self, parameters:list[int]):
        self.parameters = parameters

    def get_genom(self) -> list[int]:
        return self.parameters

    def set_health(self, health:int):
        self.health = health

    def get_health(self) -> int:
        return self.health

    def get_age(self) -> int:
        return self.age

    def get_max_age(self) -> int:
        return self.max_age

    def increase_age(self):
        self.age += 1

class Population():
    def __init__(self, number_of_individuals:int, length_of_genom:int):
        if number_of_individuals < 10:
            raise Exception("Error: Population too small.")

        self.number_of_individuals = number_of_individuals
        self.length_of_genom = length_of_genom

        self.population = [None] * self.number_of_individuals
        for i in range(self.number_of_individuals):
            self.population[i] = self.generate_random_individual()

    def get_number_of_individuals(self) -> int:
        return self.number_of_individuals

    def get_length_of_genom(self) -> int:
        return self.length_of_genom

    def generate_random_individual(self) -> Individual:
        genom = [0] * self.length_of_genom
        for i in range(self.length_of_genom):
            genom[i] = random.uniform(0, 1)
        return Individual(genom)

    def get_individuals(self) -> list[Individual]:
        return self.population

    def set_individuals(self, individuals:list[Individual]):
        self.population = individuals

    def sort_by_health(self):
        self.population = sorted(self.population, key=lambda ind: ind.get_health())

    def update_health(self, health_function):
        for ind in self.population:
            ind.set_health(health_function(ind.get_genom()))
            #print(ind.get_health())

    def breed(self, i:int, j:int) -> Individual:
        geni = self.population[i].get_genom().copy()
        genj = self.population[j].get_genom().copy()
        for k in range(len(geni)):
            select = random.randint(0, 1)
            if select == 1:
                geni[k] = genj[k]
        return Individual(geni)

    def mutate(self, i:int) -> Individual:
        gene = random.randint(0, self.length_of_genom - 1)
        self.population[i].get_genom()[gene] = random.uniform(0, 1)

    def replace(self, i:int, individual:Individual):
        self.population[i] = individual

    def replace_with_random(self, i:int):
        self.population[i] = self.generate_random_individual()

    def kill_and_replace(self):
        for i in range(len(self.population)):
            if self.population[i].get_age() > self.population[i].get_max_age():
                self.population[i] = self.generate_random_individual()

    def increase_age_of_population(self):
        for i in range(len(self.population)):
            self.population[i].increase_age()

class GeneticAlgorithm:
    def __init__(self, parameterspace:Parameterspace, population:Population, health_malus_factor:float = 1):
        self.parameterspace = parameterspace
        self.population = population
        self.health_malus_factor = health_malus_factor

    def get_bins_from_genom(self, parameters:list[int]) -> list[int]:
        bins = [int(x * self.parameterspace.get_cardinality()) for x in parameters]
        return bins

    def health_function(self, parameters:list[int]) -> int:
        bins = self.get_bins_from_genom(parameters)
        """
        Calculates the health of a individual:

            health = Sum(weight_i) - factor * malus * Sum(weight_i)
            health = (1 - factor * malus) * Sum(weight_i)
        """
        # sum up the weights in each bin
        health = 0
        weights = self.parameterspace.get_weights()
        for b in bins:
            health += weights[b]

        # for every bin that occurs more than once, reduce health
        # (forces the algorithm to look into separate bins)
        # malus can maximal be the number of genoms (=bins),
        # therefore it is reasonable to choose health_malus_factor < (1 / number_of_bins)
        counts = {}
        for item in bins:
            counts[item] = counts.get(item, 0) + 1
        malus = 0
        for c in counts:
            if counts[c] > 1:
                malus += counts[c] - 1
        health -= (self.health_malus_factor * malus * health)

        return health

    def get_max_health(self) -> int:
        genom_length = self.population.get_length_of_genom()
        return max(self.parameterspace.get_weights()) * genom_length

    def step(self) -> list[Individual]:
        self.population.increase_age_of_population()
        self.population.update_health(self.health_function)
        self.population.sort_by_health()

        # Step 1: breeding
        number_of_individuals = self.population.get_number_of_individuals()
        child12 = self.population.breed(number_of_individuals - 1, number_of_individuals - 2)
        child21 = self.population.breed(number_of_individuals - 2, number_of_individuals - 1)
        child34 = self.population.breed(number_of_individuals - 3, number_of_individuals - 4)
        child43 = self.population.breed(number_of_individuals - 4, number_of_individuals - 3)
        # replace the individuals with bad health
        self.population.replace(0, child12)
        self.population.replace(1, child21)
        self.population.replace(2, child34)
        self.population.replace(3, child43)

        # Step 2: mutate
        self.population.mutate(0)
        self.population.mutate(2)

        # Step 3: replace weak individuals with random ones
        self.population.replace_with_random(4)
        self.population.replace_with_random(5)

        # Step 4: Kill old individuals
        self.population.kill_and_replace()

        self.population.update_health(self.health_function)
        self.population.sort_by_health()

        return self.population.get_individuals()

    def run(self, threshold:float) -> list[int]:
        maxhealth = self.get_max_health()
        while True:
            individuals = self.step()
            for ind in individuals:
                print(f"health of ind {ind}: {ind.get_health()}")
            print()
            if individuals[0].get_health() >= maxhealth * threshold:
                parameters = individuals[0].get_genom()
                return self.get_bins_from_genom(parameters)

    def get_population(self) -> Population:
        return self.population

    def get_parameterspace(self) -> Parameterspace:
        return self.parameterspace

class OptimizationController():
    """
    Wrapper class for initializing parameter space binning and using the genetic algorithm to search for optimal paramters.

    Methods:
        __init__: Constructor of the OptimizationController. Parameter boundaries and parameter divisions must be provided. Parameters for the genetic algorithm are optional.
        print_best_performing_bins: Output the best performing parameter space bins. In these bins a successful glitch is assumed.
        step: Perform the next step of the genetic algorithm; should be called before every experiment.
        add_experiment: Add the parameters and the outcome (weights) to the parameter space.
    """

    def __init__(self, parameter_boundaries:list[tuple[int, int]], parameter_divisions:list[int], number_of_individuals:int = 10, length_of_genom:int = 20, malus_factor_for_equal_bins:float = 1):
        """
        Constructor of the OptimizationController. Parameter boundaries and parameter divisions must be provided. Parameters for the genetic algorithm are optional.

        Parameters:
            parameter_boundaries: Boundaries of all parameters to look into, for example `[(s_delay, e_delay), (s_t1, e_t1), (s_length, e_length)]`.
            parameter_divisions: Number of how many bins the individual parameter ranges should be divided into, for example `[10, 10, 5]`. Bigger values mean division into smaller bins which increases accuracy at the expense of performance (longer glitching campaign duration).
            number_of_individuals: Number of individuals used simultaneously in the genetic algorithm. Increasing this number improves accuracy at the expense of performance (longer glitching campaign duration).
            length_of_genom: Number of bins to track per individual. The larger the parameter space, the higher this number should be.
            malus_factor_for_equal_bins: Can be used to punish individuals that look into the same bin multiple times. Must be in range [0, 1]. A bigger value means a greater penalty.
        """
        self.par = Parameterspace(parameter_boundaries, parameter_divisions)
        self.pop = Population(number_of_individuals, length_of_genom)
        # renorm the malus factor to the maximum number of genoms,
        # thus malus_factor_for_equal_bins can be chosen between 0 and 1.
        if malus_factor_for_equal_bins > 1 or malus_factor_for_equal_bins < 0:
            print("[-] Error: malus_factor_for_equal_bins must be in range [0, 1].")
            sys.exit(-1)
        factor = malus_factor_for_equal_bins / length_of_genom
        self.opt = GeneticAlgorithm(self.par, self.pop, factor)
        self.i_current_individual = 0
        self.i_current_bin = 0
        self.number_of_individuals = self.pop.get_number_of_individuals()
        self.length_of_genom = self.pop.get_length_of_genom()

    def print_best_performing_bins(self):
        """
        Method that prints an overview of the best performing bins of the parameterspace. Can be called from time to time to obtain an overview of the optimization process.
        """
        number_of_individuals = self.pop.get_number_of_individuals()
        ind0 = self.pop.get_individuals()[number_of_individuals - 1]
        print(f"[+] Individual health = {ind0.get_health()}, age = {ind0.get_age()}")
        print("[+] Best performing bins:")
        genom = ind0.get_genom()
        bins = self.opt.get_bins_from_genom(genom)
        boundaries = []
        for i in range(len(bins)):
            boundary = self.par.get_boundaries(bins[i])
            print(f"    bin = {bins[i]}: {boundary}")
            boundaries.append(boundary)

    def step(self) -> list[int]:
        """
        Perform the next step in the optimization process. Should be called before the next experiment.

        Returns:
            A list of the parameters for the next experiment.
        """
        individuals = self.pop.get_individuals()
        parameters = individuals[self.i_current_individual].get_genom()
        bins = self.opt.get_bins_from_genom(parameters)
        boundaries = self.par.get_boundaries(bins[self.i_current_bin])

        random_numbers = []
        for b in boundaries:
            random_numbers.append(random.uniform(b[0], b[1]))

        # next bin
        self.i_current_bin += 1
        if self.i_current_bin >= self.length_of_genom:
            self.i_current_bin = 0
            # next individual
            self.i_current_individual += 1
            if self.i_current_individual >= self.number_of_individuals:
                self.i_current_individual = 0
                # next age step
                self.pop.set_individuals(self.opt.step())

        return [round(x, 1) for x in random_numbers]

    def add_experiment(self, weight:int, *parameter:int):
        """
        Method to add the parameters and the outcome (weights) to the parameter space. Can be used similarly to the method `Database.insert()`.

        Parameters:
            weight: Weight of the outcome of the experiment. Higher values mean a better performing bin and a better health of the individuals that track that bin.
            parameter: list of parameters that were used for the experiment.
        """
        self.par.add_experiment(weight, *parameter)



#!/usr/bin/env python3
# Copyright (C) 2024 Dr. Matthias Kesenheimer - All Rights Reserved.
# This file is based on TAoFI-FaultLib which is released under the GPL3 license.
# Go to https://github.com/raelize/TAoFI-FaultLib/LICENSE for full license details.
#
# You should have received a copy of the GPL3 license with this file.
# If not, please write to: m.kesenheimer@gmx.net.

"""
Findus is a Python library to perform fault-injection attacks on embedded devices. It was developed for the Pico Glitcher, however, the ChipWhisperer Pro and the ChipWhisperer Husky is also supported.

This is the documentation of the findus module and all its classes.
"""

import sqlite3
import time
import ast
import serial
import sys
import datetime
import os
import glob
import random
from findus import pyboard
from importlib.metadata import version

class Database():
    """
    Database class managing access to the SQLite databases to store results from a glitching campaign.
    The parameter points stored in these databases are used to render an overview of the scanned parameter point via a web application.
    The web application and data analyzer can be run separately from the glitcher scripts by the following command:

        analyzer --directory databases

    Example usage:

        # import Database from findus
        from findus import Database
        ...
        database = Database(argv=argv)
        ...
        database.insert(experiment_id, delay, length, color, response)

    Example usage for multi-dimensional parameter space:

        # import Database from findus
        from findus import Database
        ...
        database = Database(argv=argv, column_names=["delay", "length", "t1"])
        ...
        database.insert(experiment_id, delay, length, t1, color, response)

    If you want to plot the `(length, t1)`-slice, you can call the `analyzer` script as follows:

        analyzer --directory databases -x length -y t1

    If `dbname` is not provided, a name will automatically generated based on `argv`.

    Methods:
        __init__: Default constructor.
        insert: Method to insert datapoints into the SQLite database.
        get_parameters_of_experiment: Get the parameters of a dataset by experiment_id.
        remove: Remove a parameter point from the database by experiment_id.
        cleanup: Remove all parameter points with a given color.
        get_number_of_experiments: Get the total number of performed experiments (number of datasets in the database).
        get_latest_experiment_id: Get the latest experiment_id.
        get_base_experiments_count: [Deprecated] Get the total number of performed experiments (number of datasets in the database).
        close: Close the connection to the database.
    """

    def __init__(self, argv: list[str], dbname: str = None, resume: bool = False, nostore: bool = False, column_names = None):
        """
        Default constructor of the Database class.

        Parameters:
            argv: Arguments that were supplied to the main-script. These arguments are stored as metadata when the database is instantiated.
            dbname: Name of the database to be generated.
            resume: Resume a previous run and write the results into the previously generated database
            nostore: Do not store the results in a database (can be used for debugging).
        """
        self.nostore = nostore
        if not os.path.isdir('databases'):
            os.mkdir("databases")

        if resume and dbname is None:
            list_of_files = glob.glob('databases/*.sqlite')
            latest_file = max(list_of_files, key=os.path.getctime)[10:]
            print(f"[+] Resuming previous database {latest_file}")
            self.dbname = latest_file
        elif dbname is None:
            script_name = os.path.basename(sys.argv[0])
            self.dbname = f"{script_name}_%s.sqlite" % datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
        else:
            self.dbname = dbname

        self.con = sqlite3.connect("databases/" + self.dbname)
        self.cur = self.con.cursor()
        self.argv = argv
        self.column_names = column_names
        if column_names is None:
            self.column_names = ["delay", "length"]
        if not resume and dbname is None:
            columns = ", ".join(f"{col} integer" for col in self.column_names)
            self.cur.execute(f"CREATE TABLE experiments(id integer, {columns}, color text, response blob)")
            self.cur.execute("CREATE TABLE metadata (stime_seconds integer, argv blob)")

        self.base_row_count = self.get_latest_experiment_id()
        if self.base_row_count is None:
            self.base_row_count = 0
        if resume or dbname is not None:
            print(f"[+] Number of experiments in previous database: {self.base_row_count}")

    def insert(self, *dataset):
        """
        Method to insert datapoints into the SQLite database.

        Parameters:
            dataset: Dataset consisting of experiment_id, delay, length, [additional parameters, ...], color and response.
                - experiment_id: ID of the experiment to insert into the database.
                - delay: Time from trigger until the glitch is set (in nano seconds).
                - length: Length of glitch (in nano seconds).
                - color: Color with which the parameter point (delay, length) is to be displayed in the graph.
                - response: Byte string of target response. 
        """
        if len(dataset) < 4:
            raise Exception("Database.insert: Too less arguments given.")
        experiment_id = dataset[0]
        if not self.nostore:
            if (experiment_id + self.base_row_count) == 0:
                s_argv = ' '.join(self.argv[1:])
                self.cur.execute("INSERT INTO metadata (stime_seconds,argv) VALUES (?,?)", [int(time.time()), s_argv])
            parameters = dataset[1:-2]
            color = dataset[-2]
            response = dataset[-1]
            #print(f"{experiment_id}, {parameters}, {color}, {response}")
            columns = ",".join(f"{col}" for col in self.column_names)
            values = [experiment_id + self.base_row_count, color, response]
            # insert the parameters after position 1
            values[1:1] = parameters
            #print(values)
            qmarks = ",".join("?" for _ in range(len(values)))
            #print(qmarks)
            #print(f"INSERT INTO experiments (id,{columns},color,response) VALUES ({qmarks})")
            self.cur.execute(f"INSERT INTO experiments (id,{columns},color,response) VALUES ({qmarks})", values)
            self.con.commit()

    def get_parameters_of_experiment(self, experiment_id: int) -> list:
        """
        Get the parameters of a dataset by experiment_id.

        Parameters:
            experiment_id: ID of the experiment to insert into the database.

        Returns:
            List of parameters.
        """
        self.cur.execute("SELECT * FROM experiments WHERE id = (?);", [experiment_id + self.base_row_count])
        self.con.commit()
        return next(self.cur, [None])

    def remove(self, experiment_id: int):
        """
        Remove a parameter point from the database by experiment_id.

        Parameters:
            experiment_id: ID of the experiment to insert into the database.
        """
        self.cur.execute("DELETE FROM experiments WHERE id = (?);", [experiment_id])
        self.con.commit()

    def remove_rel(self, experiment_id: int):
        """
        Remove a parameter point from the database by experiment_id relative to the base row count.

        Parameters:
            experiment_id: ID of the experiment to insert into the database.
        """
        self.cur.execute("DELETE FROM experiments WHERE id = (?);", [experiment_id + self.base_row_count])
        self.con.commit()

    def cleanup(self, color):
        """
        Remove all parameter points with a given color.
        """
        self.cur.execute("DELETE FROM experiments WHERE color = (?);", [color])
        #self.cur.execute("DELETE FROM experiments WHERE length >= (?);", [color])
        self.con.commit()

    def get_number_of_experiments(self) -> int:
        """
        Get the total number of performed experiments (number of datasets in the database).

        Returns:
            Number of experiments performed so far in the current database.
        """
        self.cur.execute("SELECT count(id) FROM experiments")
        result = self.cur.fetchone()
        row_count = result[0]
        return row_count

    def get_latest_experiment_id(self) -> int:
        """
        Get the latest experiment_id.

        Returns:
            Experiment ID.
        """
        self.cur.execute("SELECT * FROM experiments WHERE id=(SELECT max(id) FROM experiments);")
        self.con.commit()
        return next(self.cur, [None])[0]

    def get_base_experiments_count(self) -> int:
        """
        [Deprecated] Get the total number of performed experiments (number of datasets in the database).

        Returns:
            Number of experiments performed so far in the current database.
        """
        return self.base_row_count

    def close(self):
        """
        Close the connection to the database.
        """
        self.con.close()

class Serial():
    r"""
    Class to manage serial connections more easily.
    Example usage:

        # import Serial from findus
        from findus import Serial
        ser = Serial(port="/dev/ttyUSB0")
        ser.write(b'\x7f')
        ...
        result = ser.read(3)

    Methods:
        __init__: Default constructor.
        write: Write data out via the serial interface.
        read: Read data from the serial interface.
        reset: Reset target via DTR pin and flush data lines.
        flush: Flush data buffers.
        flush_v2: Flush serial data buffers with timeout.
        close: Close serial connection.
    """
    def __init__(self, port:str = "/dev/ttyUSB0", baudrate:int = 115200, timeout:float = 0.1, bytesize:int = 8, parity:str = 'N', stopbits:int = 1):
        """
        __init__: Default constructor.

        Parameters:
            port: Port identifier of the serial connection.
            baudrate: Baudrate of the serial connection.
            timeout: Timeout after the serial connection stops listening.
            bytesize: Number of bytes per payload.
            parity: Even ('E') or Odd ('O') parity.
            stopbits: Number of stop bits.
        """
        self.ser = None
        self.port = port
        self.baudrate = baudrate
        self.timeout = timeout
        self.bytesize = bytesize
        self.parity = parity
        self.stopbits = stopbits
        self.init()

    def init(self):
        """
        Initializes the serial connection. Can be called again, if the connection was closed previously.
        """
        try:
            self.ser = serial.Serial(port=self.port, baudrate=self.baudrate, timeout=self.timeout, bytesize=self.bytesize, parity=self.parity, stopbits=self.stopbits)
        except Exception as _:
            print("[-] Serial device not found. Aborting")
            sys.exit(-1)

    def write(self, message: bytes) -> int:
        """
        Write the bytes data to the port. This should be of type `bytes` (or compatible such as `bytearray` or `memoryview`). Unicode strings must be encoded (e.g. `'hello'.encode('utf-8')`.

        Parameters:
            message: Data to send
        Returns:
            Number of bytes written.
        """
        return self.ser.write(message)

    def read(self, length: int) -> bytes:
        """
        Read `length` bytes from the serial port. If a timeout is set it may return fewer characters than requested. With no timeout it will block until the requested number of bytes is read.

        Parameters:
            length: Number of bytes to read.
        Returns:
            Bytes read from the port.
        """
        response = self.ser.read(length)
        return response
    
    def readline(self) -> bytes:
        r"""
        Read up to one line, including the \n at the end.

        Returns:
            The line received from the target.
        """
        response = self.ser.readline()
        return response

    def read_until(self, expected:bytes = '\n', size:int = None) -> bytes:
        r"""
        Read until an expected sequence is found (`\n` by default), the size is exceeded or until timeout occurs. If a timeout is set it may return fewer characters than requested. With no timeout it will block until the requested number of bytes is read.

        Parameters:
            expected: The byte string to search for.
            size: Number of bytes to read

        Returns:
            The line received from the target.
        """
        response = self.ser.read_until(expected, size)
        return response

    def reset_target(self, debug:bool = False) -> bool:
        """
        Reset target via DTR pin and flush data lines. Can be used alternatively to the reset lines of the Pico Glitcher (or ChipWhisperer Husky, or ChipWhisperer Pro) to reset the target.

        Parameters:
            debug: If set to true, garbage on the serial interface is plotted to tty.
        """
        print("[+] Resetting target...")
        self.ser.dtr = True
        time.sleep(0.1)
        self.ser.dtr = False
        response = self.ser.read(4096)
        if debug:
            for line in response.splitlines():
                print('\t', line.decode())

    def reset(self, debug:bool = False) -> bool:
        """
        Reset target via DTR pin and flush data lines. Can be used alternatively to the reset lines of the Pico Glitcher (or ChipWhisperer Husky, or ChipWhisperer Pro) to reset the target. This function is the same as `reset_target`. Left in for downward compatibility.

        Parameters:
            debug: If set to true, garbage on the serial interface is plotted to tty.
        """
        return self.reset_target(debug)

    def flush(self):
        """
        Flush serial data buffers.
        """
        self.ser.reset_input_buffer()

    def flush_v2(self, timeout:float = 0.01):
        """
        Flush serial data buffers with timeout.

        Parameters:
            timeout: Timeout after the serial connection stops listening. Can be tweaked to make sure all data is flushed.
        """
        self.ser.timeout = timeout
        self.ser.read(8192)
        self.ser.timeout = self.timeout

    def close(self):
        """
        Close serial connection.
        """
        self.ser.close()

class MicroPythonScript():
    def __init__(self, port:str = '/dev/ttyACM1', debug:bool = False):
        self.port   = None
        self.pyb    = None
        self.debug = debug

    def init(self, port:str, micropy_script:str, debug:bool = False):
        self.port = port
        try:
            self.pyb = pyboard.Pyboard(self.port)
        except Exception as _:
            print("[-] Pico Glitcher not found. Aborting")
            sys.exit(-1)
        self.pyb.enter_raw_repl()
        self.pyb.exec(f'import {micropy_script}')
        self.pyb.exec(f'mp = {micropy_script}.{micropy_script}()')

# inherit functionality and overwrite some functions
class PicoGlitcherInterface(MicroPythonScript):
    def clean_init(self):
        self.pyb.exec('mp.init()')

    def switch_pio(self, pio_base:int):
        if pio_base == 0 or pio_base == 1:
            self.pyb.exec(f'mp.switch_pio({pio_base})')

    def get_firmware_version(self):
        version_bytes = self.pyb.exec('mp.get_firmware_version()')
        decoded_str = version_bytes.decode('utf-8').strip()
        return ast.literal_eval(decoded_str)

    def set_trigger(self, mode:str, pin_trigger:str, edge_type:str = "rising"):
        self.pyb.exec(f'mp.set_trigger("{mode}", "{pin_trigger}", "{edge_type}")')

    def set_number_of_edges(self, number_of_edges:int):
        self.pyb.exec(f'mp.set_number_of_edges({number_of_edges})')

    def set_frequency(self, frequency:int):
        self.pyb.exec(f'mp.set_frequency({frequency})')

    def get_frequency(self):
        return self.pyb.exec('mp.get_frequency()')

    def set_baudrate(self, baud:int):
        self.pyb.exec(f'mp.set_baudrate({baud})')

    def set_number_of_bits(self, number_of_bits:int):
        self.pyb.exec(f'mp.set_number_of_bits({number_of_bits})')

    def set_pattern_match(self, pattern:int):
        self.pyb.exec(f'mp.set_pattern_match({pattern})')

    def power_cycle_target(self, power_cycle_time:float):
        self.pyb.exec(f'mp.power_cycle_target({power_cycle_time})')

    def power_cycle_reset(self, power_cycle_time:float):
        self.pyb.exec(f'mp.power_cycle_reset({power_cycle_time})')

    def set_mux_voltage(self, voltage:str):
        self.pyb.exec(f'mp.set_mux_voltage("{voltage}")')

    def cleanup_pio(self):
        self.pyb.exec('mp.cleanup_pio()')

    def arm(self, delay:int, length:int, number_of_pulses:int = 1, delay_between:int = 0):
        if number_of_pulses == 1:
            self.pyb.exec(f'mp.arm({delay}, {length})')
        else:
            self.pyb.exec(f'mp.arm({delay}, {length}, {number_of_pulses}, {delay_between})')

    def arm_multiplexing(self, delay:int, mul_config:dict, vinit:str = "config"):
        return self.pyb.exec(f'mp.arm_multiplexing({delay}, {mul_config}, "{vinit}")')

    def arm_pulseshaping_from_config(self, delay:int, ps_config:list[list[int]]):
        return self.pyb.exec(f'mp.arm_pulseshaping_from_config({delay}, {ps_config})')

    def arm_pulseshaping_from_spline(self, delay:int, xpoints:list[int], ypoints:list[float]):
        return self.pyb.exec(f'mp.arm_pulseshaping_from_spline({delay}, {xpoints}, {ypoints})')

    def arm_pulseshaping_from_lambda(self, delay:int, ps_lambda:str, pulse_number_of_points:int):
        return self.pyb.exec(f'mp.arm_pulseshaping_from_lambda({delay}, {ps_lambda}, {pulse_number_of_points})')

    def arm_pulseshaping_from_list(self, delay:int, pulse:list[int]):
        return self.pyb.exec(f'mp.arm_pulseshaping_from_list({delay}, {pulse})')

    def initiate_reset(self):
        self.pyb.exec('mp.initiate_reset()')

    def release_reset(self):
        self.pyb.exec('mp.release_reset()')

    def disable_vtarget(self):
        self.pyb.exec('mp.disable_vtarget()')

    def enable_vtarget(self):
        self.pyb.exec('mp.enable_vtarget()')

    def reset_target(self, reset_time:float):
        self.pyb.exec(f'mp.reset_target({reset_time})')

    def configure_gpio_out(self, pin_number:int):
        self.pyb.exec(f'mp.configure_gpio_out({pin_number})')

    def set_gpio(self, pin_number:int, value:int):
        self.pyb.exec(f'mp.set_gpio({pin_number}, {value})')

    def block(self, timeout:float):
        self.pyb.exec(f'mp.block({timeout})')

    def get_sm1_output(self) -> str:
        return self.pyb.exec('mp.get_sm1_output()')

    def set_lpglitch(self):
        self.pyb.exec('mp.set_lpglitch()')

    def set_hpglitch(self):
        self.pyb.exec('mp.set_hpglitch()')

    def set_multiplexing(self):
        self.pyb.exec('mp.set_multiplexing()')

    def set_pulseshaping(self, vinit=1.8):
        self.pyb.exec(f'mp.set_pulseshaping({vinit})')

    def do_calibration(self, vhigh:float):
        self.pyb.exec(f'mp.do_calibration({vhigh})')

    def apply_calibration(self, vhigh:float, vlow:float, store:bool = True):
        self.pyb.exec(f'mp.apply_calibration({vhigh}, {vlow}, {store})')

    def waveform_generator(self, frequency:int, gain:float, waveid:int):
        return self.pyb.exec(f'mp.waveform_generator({frequency}, {gain}, {waveid})')

    def set_dead_zone(self, dead_time:float, pin_condition:str, condition:str):
        self.pyb.exec(f'mp.set_dead_zone({dead_time}, "{pin_condition}", "{condition}")')

    def change_config_and_reset(self, key, value) -> str:
        return self.pyb.exec(f'mp.change_config_and_reset("{key}", "{value}")')

    def change_config_and_apply(self, key, value) -> str:
        return self.pyb.exec(f'mp.change_config_and_apply("{key}", "{value}")')

    def arm_adc(self):
        self.pyb.exec('mp.arm_adc()')

    def get_adc_samples(self, timeout:float = 1.0) -> list[int]:
        return self.pyb.exec(f'mp.get_adc_samples({timeout})')

    def configure_adc(self, number_of_samples:int = 1024, sampling_freq:int = 500_000):
        self.pyb.exec(f'mp.configure_adc({number_of_samples}, {sampling_freq})')

    def stop_core1(self):
        self.pyb.exec('mp.stop_core1()')

    def hard_reset(self):
        self.pyb.exec('mp.hard_reset()')

class Glitcher():
    """
    Glitcher template class. This class defines a common anchestor from which other glitcher modules should inherit from.

    Methods:
        __init__: Default constructor. Does nothing in this case.
        classify: Template method to classify an output state.
        colorize: Returns a colored string depending on a color identifier (G, Y, R, M, C, B).
        get_speed: Calculate and return the average speed of the glitching campaign (glitches per second).
    """
    def __init__(self):
        """
        Default constructor. Does nothing in this case.
        """
        pass

    def classify(self, state:bytes) -> str:
        """
        Template method to classify an output state. Overload this class if you want to customize the targets response classification. Alternatively, use the built-in class `GlitchState` to characterize the targets responses. Remember to define certain response states depending on the possible responses. See class `STM32Bootloader` for an example.

            from findus import PicoGlitcher
            from findus.STM32Bootloader import STM32Bootloader
            glitcher = PicoGlitcher()
            ...
            programmer = STM8Programmer(port=self.args.target, baud=115200)
            ...
            state = programmer.bootloader_enter()
            ...
            glitcher.classify(state)
        """
        color = 'C'
        if b'expected' in state:
            color = 'G'
        elif b'error' in state:
            color = 'M'
        elif b'warning' in state:
            color = 'O'
        elif b'timeout' in state:
            color = 'Y'
        elif b'ok' in state:
            color = 'C'
        elif b'success' in state:
            color = 'R'
        return color

    def colorize(self, s:str, color:str) -> str:
        """
        Returns a colorized string depending on a color identifier (G, Y, R, M, C, B, O, Z).
        
        Parameters:
            s: The string you want to colorize.
            color: Color identifier, one of 'G', 'Y', 'R', 'M', 'C', 'B', 'O', 'Z' (black).
        Returns:
            Returns the colorized string.
        """
        colors = {
            'G': [182, 214, 168],
            'Y': [232, 237, 164],
            'R': [228, 145, 167],
            'M': [234, 192, 226],
            'C': [113, 198, 177],
            'B': [109, 174, 217],
            'O': [197, 148, 124],
            'Z': [31, 31, 31]
        }
        #colors = {
        #    'G': [0, 255, 0],
        #    'Y': [255, 255, 0],
        #    'R': [170, 0, 0],
        #    'M': [255, 0, 255],
        #    'C': [0, 255, 255],
        #    'B': [0, 0, 255],
        #    'O': [255, 100, 50],
        #    'Z': [0, 0, 0]
        #}
        r = colors[color][0]
        g = colors[color][1]
        b = colors[color][2]
        return f"\033[38;2;{r};{g};{b}m{s}\033[0m"

    def get_speed(self, start_time:int, number_of_experiments:int) -> int:
        """
        Calculate and return the average speed of the glitching campaign (glitches per second).
        
        Parameters:
            start_time: Start time of the glitching campaign in seconds.
            number_of_experiments: Number of experiments carried out so far.
        Returns:
            Returns the average number of experiments per second.
        """
        elapsed_time = int(time.time()) - start_time
        if elapsed_time == 0:
            return 'NA'
        else:
            return number_of_experiments // elapsed_time

class PicoGlitcher(Glitcher):
    """
    Class giving access to the functions of the Pico Glitcher. Derived from Glitcher class.
    For an example, connect the Pico Glitcher as follows:

    - Remove any capacitors on your target device that could infere with the glitch.
    - Set the desired output voltage `VTARGET` with the micro switch.
    - Connect `VTARGET` with the voltage input of your target (`VCC`).
    - Connect the `GLITCH` output (either the SMA connector or the pin header) to an appropriate target pin, for example `VCC`.
    - Connect the `RESET` output with the target's reset input.
    - Connect the `RESET` line with the `TRIGGER` input.
    
    Code snippet:

        from findus import PicoGlitcher
        glitcher = PicoGlitcher()
        glitcher.init(port="/dev/ttyACM0", ext_power="/dev/ttyACM1", ext_power_voltage=3.3)
        # set up database, define delay and length
        ...
        # one shot glitching
        glitcher.arm(delay, length)
        # reset target for 0.01 seconds (the rising edge on reset line triggers the glitch)
        glitcher.reset_target(0.01)
        # read the response from the device (for example UART, SWD, etc.)
        response = ...
        # classify the response and put into database
        color = glitcher.classify(response)
        database.insert(experiment_id, delay, length, color, response)

    Methods:
        __init__: Default constructor. Does nothing in this case.
        init: Default initialization procedure.
        arm: Arm the Pico Glitcher and wait for trigger condition.
        block: Block the main script until trigger condition is met. Times out.
        reset: Reset the target via the Pico Glitcher's `RESET` output.
        release_reset: Release the reset to the target via the Pico Glitcher's `RESET` output.
        power_cycle_target: Power cycle the target via the Pico Glitcher `VTARGET` output or optionally via the multiplexing stage.
        power_cycle_reset: Power cycle and reset the target via the Pico Glitcher `RESET` and `VTARGET` output.
        reset_and_eat_it_all: Reset the target and flush the serial buffers.
        reset_wait: Reset the target and read from serial.
        set_lpglitch: Enable low-power MOSFET for glitch generation.
        set_hpglitch: Enable high-power MOSFET for glitch generation.
        rising_edge_trigger: Configure the Pico Glitcher to trigger on a rising edge on the `TRIGGER` line.
        uart_trigger: Configure the Pico Glitcher to trigger when a specific byte pattern is observed on the `TRIGGER` line.
        set_cpu_frequency: Set the CPU frequency of the Raspberry Pi Pico.
        get_cpu_frequency: Get the current CPU frequency of the Raspberry Pi Pico.
        __del__: Default deconstructor. Disconnects Pico Glitcher.
    """
    def __init__(self):
        """
        Default constructor. Does nothing in this case.
        """
        self.pico_glitcher = None

    def __del__(self):
        print("[+] Terminating gracefully.")
        try:
            self.stop_core1()
            self.hard_reset()
        except Exception as _:
            pass

    def clean_init(self):
        """
        Re-initialize the Pico Glitcher on the fly. Can be used to make changes to the configuration on the fly.
        """
        self.pico_glitcher.clean_init()

    def init(self, port:str, ext_power:str = None, ext_power_voltage:float = 3.3, enable_vtarget:bool = True):
        """
        Default initialization procedure of the Pico Glitcher. Default configuration is:

        - Set the trigger input to rising-edge trigger on `TRIGGER` input and assume triggering when the reset is released.
        - Set a dead time that prohibits triggering within a certain time (trigger rejection). This is intended to exclude false trigger conditions.
        - Use the high-power crowbar MOSFET.

        Parameters:
            port: Port identifier of the Pico Glitcher.
            ext_power: Port identifier of the external power supply (RD6006). If None, target is assumed to be supplied by `VTARGET` of the Pico Glitcher.
            ext_power_voltage: Supply voltage of the external power supply. Must be used in combination with `ext_power`. You can not control the supply voltage `VTARGET` of the Pico Glitcher with this parameter.
        """
        self.pico_glitcher = PicoGlitcherInterface()
        self.pico_glitcher.init(port, 'PicoGlitcher')

        # check compatibility
        try:
            pg_fw_version = self.pico_glitcher.get_firmware_version()
            fi_fw_version = list(map(int, version("findus").split('.')))
            print(f"[+] Version of Pico Glitcher: {pg_fw_version}")
            print(f"[+] Version of findus: {fi_fw_version}")
            # check only major and minor version, but not the build number
            if pg_fw_version[:2] != fi_fw_version[:2]:
                raise Exception("Version mismatch")
        except Exception as _:
            print("[-] Fatal error: Versions of findus and Pico Glitcher do not match.")
            print("[*] Update the Pico Glitcher firmware and findus software. See README.md.")
            print("[*] pip install --upgrade findus")
            print("[*] cd .venv/lib/python3.xx/site-packages/findus/firmware")
            print("[*] upload --port /dev/tty.<rpi-tty-port> --files AD910X.py FastADC.py Globals.py PicoGlitcher.py PulseGenerator.py Spline.py <config-version>/config.json")
            sys.exit(-1)

        if ext_power is not None:
            from findus.ExternalPowerSupply import ExternalPowerSupply
            self.pico_glitcher.disable_vtarget()
            self.power_supply = ExternalPowerSupply(port=ext_power)
            self.power_supply.set_voltage(ext_power_voltage)
            print(self.power_supply.status())
            if enable_vtarget:
                self.power_supply.enable_vtarget()
        else:
            if enable_vtarget:
                self.pico_glitcher.enable_vtarget()
            self.power_supply = None

    def get_power_supply(self):
        """
        To control the external power supply in a script, the `power_supply` object is necessary.

        Returns:
            Returns the `power_supply` object if the external power supply is used.
        """
        return self.power_supply

    def cleanup_pio(self):
        """
        Remove all programs from the internal statemachines. Could be necessary before a program is added to the statemachine but it is out of memory. Some weird behavior was observed when removing and adding programs to the statemachines. Use with caution.
        """
        self.pico_glitcher.cleanup_pio()

    def switch_pio(self, pio_base:int):
        """
        In some rare cases, it might be necessary to switch to the other internal statemachine. For example, if a change in the glitch configuration is necessary but the statemachine is out of memory. Instead of removing the programs from the statemachine (`cleanup_pio`), one can simply switch to the other statemachine and setup the Pico Glitcher to use the configuration on the second statemachine.

        Parameters:
            pio_base: Base number of the statemachine. Can either be 0 or 1. Other values are ignored.
        """
        self.pico_glitcher.switch_pio(pio_base)

    def arm(self, delay:int, length:int, number_of_pulses:int = 1, delay_between:int = 0):
        """
        Arm the Pico Glitcher and wait for the trigger condition. The trigger condition can either be when the reset on the target is released or when a certain pattern is observed in the serial communication.

        Parameters:
            delay: Glitch is emitted after this time. Given in nano seconds. Expect a resolution of about 5 nano seconds.
            length: Length of the glitch in nano seconds. Expect a resolution of about 5 nano seconds.
            number_of_pulses: The number of pulses to emit. This can be used to emit bursts of crowbar glitches.
            delay_between: The delay between each pulse.
        """
        self.pico_glitcher.arm(delay, length, number_of_pulses, delay_between)

    def arm_multiplexing(self, delay:int, mul_config:dict, vinit:str = "config"):
        """
        Arm the Pico Glitcher and wait for the trigger condition. The trigger condition can either be when the reset on the target is released or when a certain pattern is observed in the serial communication. Only available for hardware revision 2 and later.

        Parameters:
            delay: Glitch is emitted after this time. Given in nano seconds. Expect a resolution of about 5 nano seconds.
            mul_config: The dictionary for the multiplexing profile with pairs of identifiers and values. For example, this could be `{"t1": 10, "v1": "GND", "t2": 20, "v2": "1.8", "t3": 30, "v3": "GND", "t4": 40, "v4": "1.8"}`. Meaning that when triggered, a GND-voltage pulse with duration of `10ns` is emitted, followed by a +1.8V step with duration of `20ns` and so on.
            vinit: The initial value of the multiplexer. If `"config"` is chosen, the initial value is read from the configuration file. Additionally, the user can choose between `"VI1"` or `"VI2"`.
        """
        self.pico_glitcher.arm_multiplexing(delay, mul_config, vinit)

    def arm_pulseshaping_from_config(self, delay:int, ps_config:list[list[int]]):
        """
        Arm the Pico Glitcher and wait for the trigger condition. The trigger condition can either be when the reset on the target is released or when a certain pattern is observed in the serial communication. Only available for hardware revision 2 and later. Additionally, the Pulse Shaping Expansion board is needed.

        Parameters:
            delay: Glitch is emitted after this time. Given in nano seconds. Expect a resolution of about 5 nano seconds.
            ps_config: The pulse configuration given as a list of time deltas and voltage values.

        Example:

            ps_config = [[4*length, 1.8], [4*length, 0.95], [length, 0.0]]
            glitcher.arm_pulseshaping_from_config(delay, ps_config)
        """
        return self.pico_glitcher.arm_pulseshaping_from_config(delay, ps_config)

    def arm_pulseshaping_from_spline(self, delay:int, xpoints:list[int], ypoints:list[float]):
        """
        Arm the Pico Glitcher and wait for the trigger condition. The pulse definition is given by time and voltage points. Intermediate values are interpolated.

        Parameters:
            delay: Glitch is emitted after this time. Given in nano seconds. Expect a resolution of about 5 nano seconds.
            xpoints: A list of time points (in nanoseconds) where voltage changes occur.
            ypoints: The corresponding voltage levels at each time point.

        Example:

            xpoints = [0,   100, 200, 300, 400, 500, 515, 520]
            ypoints = [3.0, 2.1, 2.0, 2.0, 1.7, 0.0, 2.0, 3.0]
            glitcher.arm_pulseshaping_from_spline(delay, xpoints, ypoints)
        """
        return self.pico_glitcher.arm_pulseshaping_from_spline(delay, xpoints, ypoints)

    def arm_pulseshaping_from_lambda(self, delay:int, ps_lambda:str, pulse_number_of_points:int):
        """
        Arm the Pico Glitcher and wait for the trigger condition. Generate the pulse from a lambda function depending on the time.

        Parameters:
            delay: Glitch is emitted after this time. Given in nano seconds. Expect a resolution of about 5 nano seconds.
            ps_lambda: A lambda function that defines the glitch at certain times. Must be given as string which is processed by the Pico Glitcher at runtime.
            pulse_number_of_points: The approximate length of the pulse. This is needed to constrain the pulse and to save computing time.

        Example:

            ps_lambda = f"lambda t:-1.0/({2*length})*t+3.0 if t<{2*length} else 2.0 if t<{4*length} else 0.0 if t<{5*length} else 3.0"
            glitcher.arm_pulseshaping_from_lambda(delay, ps_lambda, 6*length)
        """
        return self.pico_glitcher.arm_pulseshaping_from_lambda(delay, ps_lambda, pulse_number_of_points)

    def arm_pulseshaping_from_list(self, delay:int, pulse:list[int]):
        """
        Arm the Pico Glitcher and wait for the trigger condition. Genereate the pulse from a raw array of values.

        Parameters:
            delay: Glitch is emitted after this time. Given in nano seconds. Expect a resolution of about 5 nano seconds.
            pulse: A raw list of points that define the pulse. No calibration and no constraints are applied to the list. The list is forwarded directly to the DAC.

        Example:

            pulse = [-0x1fff] * 50 + [-0x0fff] * 50 + [-0x07ff] * 50 + [0x0000] * 50
            glitcher.arm_pulseshaping_from_list(delay, pulse)
        """
        return self.pico_glitcher.arm_pulseshaping_from_list(delay, pulse)

    def block(self, timeout:float = 1.0):
        """
        Block until trigger condition is met. Raises an exception if times out.
        
        Parameters:
            timeout: Time after the block is released.
        """
        self.pico_glitcher.block(timeout)

    def get_sm1_output(self) -> str:
        return self.pico_glitcher.get_sm1_output()

    def reset_target(self, reset_time:float = 0.2):
        """
        Reset the target via the Pico Glitcher's `RESET` output.

        Parameters:
            reset_time: Time how long the target is held in reset.
        """
        self.pico_glitcher.reset_target(reset_time)

    def reset(self, reset_time:float = 0.2):
        """
        Reset the target via the Pico Glitcher's `RESET` output. This function is the same as `reset_target`. Left in for downward compatibility.
        
        Parameters:
            reset_time: Time how long the target is held in reset.
        """
        self.pico_glitcher.reset_target(reset_time)

    def set_gpio(self, pin_number:int, value:int):
        """
        Set the GPIO pin `pin_number` to a specific output value (0 or 1).

        Parameters:
            pin_number: GPIO pin number (for example 4, 5, 6).
            value: Output value of the GPIO pin (0 or 1).
        """
        self.pico_glitcher.set_gpio(pin_number, value)

    def configure_gpio_out(self, pin_number:int):
        """
        Configure the GPIO pin `pin_number` as an output.

        Parameters:
            pin_number: GPIO pin number (for example 4, 5, 6).
        """
        self.pico_glitcher.configure_gpio_out(pin_number)

    def release_reset(self):
        """
        Release the reset of the target via the Pico Glitcher's `RESET` output.
        """
        self.pico_glitcher.release_reset()

    def initiate_reset(self):
        """
        Initiate the reset of the target via the Pico Glitcher's `RESET` output.
        """
        self.pico_glitcher.initiate_reset()

    def enable_vtarget(self):
        """
        Enable the Pico Glitcher's `VTARGET` output.
        """
        self.pico_glitcher.enable_vtarget()

    def disable_vtarget(self):
        """
        Enable the Pico Glitcher's `VTARGET` output.
        """
        self.pico_glitcher.disable_vtarget()

    def set_mux_voltage(self, voltage:str):
        """
        TODO
        """
        self.pico_glitcher.set_mux_voltage(voltage)

    def power_cycle_target(self, power_cycle_time:float = 0.2):
        """
        Power cycle the target via the Pico Glitcher `VTARGET` output or optionally via the multiplexing stage. Depending on the glitch configuration (crowbar, multiplexing, pulse-shaping), the power-cycle is performed with either the crowbar, multiplexing or pulse-shaping stage.
        If available, target is power-cycled by the external power supply RD6006.
        
        Parameters:
            power_cycle_time: Time how long the power supply is cut. If `ext_power` is defined, the external power supply (RD6006) is cycled.
        """
        if self.power_supply is not None:
            self.power_supply.power_cycle_target(power_cycle_time)
        else:
            self.pico_glitcher.power_cycle_target(power_cycle_time)

    def power_cycle_reset(self, power_cycle_time:float = 0.2):
        """
        Power cycle and reset the target via the Pico Glitcher `VTARGET` and `RESET` output. Optionally the multiplexing stage can be used to power-cycle the target. Can also be used to define sharper trigger conditions via the `RESET` line. Depending on the glitch configuration (crowbar, multiplexing, pulse-shaping), the power-cycle is performed with either the crowbar, multiplexing or pulse-shaping stage.
        
        Parameters:
            power_cycle_time: Time how long the power supply is cut. If `ext_power` is defined, the external power supply is cycled.
        """
        if self.power_supply is not None:
            self.power_supply.disable_vtarget()
            self.pico_glitcher.initiate_reset()
            time.sleep(power_cycle_time)
            self.power_supply.enable_vtarget()
            self.pico_glitcher.release_reset()
        else:
            self.pico_glitcher.power_cycle_reset(power_cycle_time)

    def reset_and_eat_it_all(self, target:serial.Serial, target_timeout:float = 0.3):
        """
        Reset the target via the Pico Glitcher's `RESET` output and flush the serial buffers.
        
        Parameters:
            target: Serial communication object (usually defined as `target = serial.Serial(...)`).
            target_timeout: Time-out of the serial communication. After this time, reading from the serial connection is canceled and it is assumed that there is no more garbage on the line.
        """
        self.pico_glitcher.initiate_reset()
        target.ser.timeout = target_timeout
        target.read(4096)
        target.ser.timeout = target.timeout
        self.pico_glitcher.release_reset()

    def reset_wait(self, target:serial.Serial, token:bytes, reset_time:float = 0.2, debug:bool = False) -> bytes:
        """
        Reset the target via the Pico Glitchers's `RESET` output and wait until the target responds (read from serial).

        Parameters:
            target: Serial communication object (usually defined as `target = serial.Serial(...)`).
            token: Expected response from target. Read from serial multiple times until target responds.
            reset_time:  Time how long the target is held under reset.
            debug: If `true`, more output is given.

        Returns:
            Returns the target's response.
        """
        self.pico_glitcher.initiate_reset()
        time.sleep(reset_time)
        self.pico_glitcher.release_reset()
        response = target.read(4096)
        for _ in range(0, 5):
            if token in response:
                break
            response += target.read(4096)
        if debug:
            for line in response.splitlines():
                print('\t', line.decode())
        return response

    def set_lpglitch(self):
        """
        Enable the low-power crowbar MOSFET for glitch generation.

        The glitch output is an SMA-connected output line that is normally connected to a target's power rails. If this setting is enabled, a low-powered MOSFET shorts the power-rail to ground when the glitch module's output is active.
        """
        self.pico_glitcher.set_lpglitch()

    def set_hpglitch(self):
        """
        Enable the high-power crowbar MOSFET for glitch generation.
        
        The glitch output is an SMA-connected output line that is normally connected to a target's power rails. If this setting is enabled, a high-powered MOSFET shorts the power-rail to ground when the glitch module's output is active.
        """
        self.pico_glitcher.set_hpglitch()

    def set_multiplexing(self):
        """
        Enables the multiplexing mode of the Pico Glitcher version 2 to quickly switch between different voltage levels.
        """
        self.pico_glitcher.set_multiplexing()

    def set_pulseshaping(self, vinit:float=1.8):
        """
        Enables the pulse-shaping mode of the Pico Glitcher version 2 to apply a voltage profile to the target's supply voltage.

        Parameters:
            vinit: The initial voltage (voltage offset) to base the calculations on. This does not change the output voltage of the pulse shaping expansion board. However, this parameter is used to calculate the correct offsets and scaling of the pulse.
        """
        self.pico_glitcher.set_pulseshaping(vinit)

    def do_calibration(self, vhigh:float):
        """
        Emit a calibration pulse with the Pico Glitcher Pulse Shaping expansion board to determine `vhigh` and `vlow`. These parameters are used to calculate the offset and gain parameters of the DAC.

        Parameters:
            vhigh: The initial voltage to perform the calibration with. Default is `1V`.
        """
        self.pico_glitcher.do_calibration(vhigh)

    def apply_calibration(self, vhigh:float, vlow:float, store:bool = True):
        """
        Calculate and store the offset and gain parameters that were determined by the calibration routine. These values are stored in `config.json` and must be re-calculated if the config is overwritten.

        Parameters:
            vhigh: The maximum voltage of the calibration voltage trace.
            vlow: The minimum voltage of the calibration voltage trace.
            store: wether to store the offset and gain factor in the Pico Glitcher configuration.
        """
        self.pico_glitcher.apply_calibration(vhigh, vlow, store)

    def tio_trigger(self, pin_trigger:str = "default", edge_type="rising"):
        """
        Configure the Pico Glitcher to trigger on a rising or falling edge on the `TRIGGER` line.

        Parameters:
            pin_trigger: The trigger pin to use. Can either be "default" (default `TRIGGER` input) or "alt" (alternative trigger input `TRIGGER1`). For hardware version 2 options "ext1" or "ext2" are also available.
            edge_type: Trigger on a "rising" (default) or "falling" edge.
        """
        self.pico_glitcher.set_trigger("tio", pin_trigger, edge_type)

    def rising_edge_trigger(self, pin_trigger:str = "default", dead_time:float = 0, pin_condition:str = "default", condition:str = "rising"):
        """
        Configure the Pico Glitcher to trigger on a rising edge on the `TRIGGER` line with optional trigger suppression (dead time).
        
        Parameters:
            pin_trigger: The trigger pin to use. Can either be "default" (default `TRIGGER` input) or "alt" (alternative trigger input `TRIGGER1`). For hardware version 2 options "ext1" or "ext2" are also available.
            dead_time: Set a dead time that prohibits triggering within a certain time (trigger rejection). This is intended to exclude false trigger conditions. Can also be set to 0 to disable this feature.
            pin_condition: The rejection time is generated internally by measuring the state of the the given pin of the Pico Glitcher. If you want to trigger on the reset condition, set `pin_condition = 'reset'`, else if you want to trigger on the target power set `pin_condition = 'power'`. `pin_condition` can either be "default", "power", "reset" or a GPIO pin number (for example "4", "5" or "6"). If `dead_time` is set to zero and `pin_condition = 'default'`, this parameter is ignored.
            condition: Can either be "falling" or "rising". The `dead_time` is measured on the pin `pin_condition` after the specified condition (falling- or rising-edge). For example, a good choice is "rising" for the "default" configuration and "falling" for the "reset" configuration. However, this could depend on the specific use case.
        """
        self.pico_glitcher.set_trigger("tio", pin_trigger, "rising")
        self.pico_glitcher.set_dead_zone(dead_time, pin_condition, condition)

    def falling_edge_trigger(self, pin_trigger:str = "default", dead_time:float = 0, pin_condition:str = "default", condition:str = "rising"):
        """
        Configure the Pico Glitcher to trigger on a falling edge on the `TRIGGER` line with optional trigger suppression (dead time).

        Parameters:
            pin_trigger: The trigger pin to use. Can either be "default" (default `TRIGGER` input) or "alt" (alternative trigger input `TRIGGER1`). For hardware version 2 options "ext1" or "ext2" are also available.
            dead_time: Set a dead time that prohibits triggering within a certain time (trigger rejection). This is intended to exclude false trigger conditions. Can also be set to 0 to disable this feature.
            pin_condition: The rejection time is generated internally by measuring the state of the the given pin of the Pico Glitcher. If you want to trigger on the reset condition, set `pin_condition = 'reset'`, else if you want to trigger on the target power set `pin_condition = 'power'`. `pin_condition` can either be "default", "power", "reset" or a GPIO pin number (for example "4", "5" or "6"). If `dead_time` is set to zero and `pin_condition = 'default'`, this parameter is ignored.
            condition: Can either be "falling" or "rising". The `dead_time` is measured on the pin `pin_condition` after the specified condition (falling- or rising-edge). For example, a good choice is "rising" for the "default" configuration and "falling" for the "reset" configuration. However, this could depend on the specific use case.
        """
        self.pico_glitcher.set_trigger("tio", pin_trigger, "falling")
        self.pico_glitcher.set_dead_zone(dead_time, pin_condition, condition)

    def uart_trigger(self, pattern:int, baudrate:int = 115200, number_of_bits:int = 8, pin_trigger:str = "default"):
        """
        Configure the Pico Glitcher to trigger when a specific byte pattern is observed on the `TRIGGER` line.
        
        Parameters:
            pattern: Byte pattern that is transmitted on the serial lines to trigger on. For example `0x11`.
            baudrate: The baudrate of the serial communication.
            number_of_bits: The number of bits of the UART payload.
            pin_trigger: The trigger pin to use. Can be either "default" or "alt". For hardware version 2 options "ext1" or "ext2" can also be chosen.
        """
        self.pico_glitcher.set_trigger("uart", pin_trigger)
        self.pico_glitcher.set_baudrate(baudrate)
        self.pico_glitcher.set_number_of_bits(number_of_bits)
        self.pico_glitcher.set_pattern_match(pattern)

    def edge_count_trigger(self, pin_trigger:str = "default", number_of_edges:int = 2, edge_type:str = "rising"):
        """
        Configure the Pico Glitcher to trigger after a certain number of eddges on the `TRIGGER` line.

        Parameters:
            pin_trigger: The trigger pin to use. Can either be "default" (default `TRIGGER` input) or "alt" (alternative trigger input `TRIGGER1`). For hardware version 2 options "ext1" or "ext2" are also available.
            number_of_edges: The number of edges after which the Pico Glitcher triggers.
            edge_type: Trigger on a "rising" (default) or "falling" edge.
        """
        self.pico_glitcher.set_trigger("edge", pin_trigger, edge_type)
        self.pico_glitcher.set_number_of_edges(number_of_edges)

    def set_cpu_frequency(self, frequency:int = 200_000_000):
        """
        Set the CPU frequency of the Raspberry Pi Pico.
        
        Parameters:
            frequency: the CPU frequency.
        """
        self.pico_glitcher.set_frequency(frequency)

    def get_cpu_frequency(self) -> int:
        """
        Get the current CPU frequency of the Raspberry Pi Pico.
        
        Returns:
            Returns the CPU frequency.
        """
        return self.pico_glitcher.get_frequency()

    def change_config_and_reset(self, key:str, value:int|float|str):
        """
        Change the content of the configuration file `config.json`. Note that the value to be changed must already exist. After calling this function, the Pico Glitcher must be re-initialized (call `glitcher.init(port=args.rpico, ext_power=args.power, ext_power_voltage=3.3)` again).

        Parameters:
            key: Key of value to be replaced.
            value: Value to be set.
        """
        return self.pico_glitcher.change_config_and_reset(key, value)

    def change_config_and_apply(self, key:str, value:int|float|str):
        """
        Change the content of the configuration file `config.json`. Note that the value to be changed must already exist. Apply the values afterwards without reset by calling `glitcher.switch_pio(1)`. This will re-initialize the internal statemachines and apply the configuration.

        Parameters:
            key: Key of value to be replaced.
            value: Value to be set.
        """
        return self.pico_glitcher.change_config_and_apply(key, value)

    def waveform_generator(self, frequency:int, gain:float, waveid:int):
        """
        The Pulse Shaping expansion board of the Pico Glitcher v2 can be used to generate arbitrary and continous waveforms as well.

        Parameters:
            frequency: The frequency of the signal.
            gain: The gain (overall amplitude) of the signal.
            waveid: This determines the signal type, that is, what signal should be generated.

        - sine wave: `waveid = 0`
        - cosine wave: `waveid = 1`
        - triangle: `waveid = 2`
        - positive sawtooth: `waveid = 3`
        - negative sawtooth: `waveid = 4`
        """
        self.pico_glitcher.waveform_generator(frequency, gain, waveid)

    def arm_adc(self):
        """
        Arm the ADC on pin 26 and capture ADC samples if the trigger condition is met. On Pico Glitcher hardware version 1, the separate SMA connector labeled `Analog` can be used to measure analog voltage traces. On revision 2, the analog input is directly connected to the `GLITCH` line.
        """
        self.pico_glitcher.arm_adc()

    def get_adc_samples(self, timeout:float = 1.0) -> list[int]:
        """
        Read back the captured ADC samples.
        """
        samples = self.pico_glitcher.get_adc_samples(timeout)
        #print(samples)
        decoded_str = samples.decode().strip()
        num_str = decoded_str.split("[")[1].split("]")[0]
        int_list = [int(x) for x in num_str.split(",")]
        return int_list

    def configure_adc(self, number_of_samples:int = 1024, sampling_freq:int = 500_000):
        """
        Configure the onboard ADC of the Pico Glitcher.

        Parameters:
            number_of_samples: The number of samples to capture after triggering.
            sampling_freq: The sampling frequency of the ADC. `500 kSPS` is the maximum for the Pico Glitcher.
        """
        self.pico_glitcher.configure_adc(number_of_samples, sampling_freq)

    def stop_core1(self):
        """
        Stop execution on the second core of the Pico Glitcher (Raspberry Pi Pico).
        """
        self.pico_glitcher.stop_core1()

    def hard_reset(self):
        """
        Perform a hard reset of the Pico Glitcher (Raspberry Pi Pico).
        """
        try:
            self.pico_glitcher.hard_reset()
        except Exception as _:
            pass

class Helper():
    """
    Helper class that provides useful functions.
    Example usage:
    
        from findus import Helper as helper
        filename = f"{helper.timestamp()}_memory_dump.bin"

    Methods:
        timestamp: Provides the current timestamp in a file-friendly format.
    """
    def timestamp() -> str:
        """
        Provides the current timestamp in a file-friendly format.
        
        Returns:
            Returns the current timestamp in the format %Y-%m-%d_%H-%M-%S.
        """
        return datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")

    def linspace(start, stop, num, endpoint=True) -> list:
        """
        Generate equidistant points from a start to a stop point. Equivalent to numpy's linspace function.

        Parameters:
            start: Start point of the interval.
            stop: End point of the interval.
            num: Number of points to divide the interval into.
            endpoint: Whether the stop point should be part of the interval.

        Returns:
            List of integer or real numbers.
        """
        if num <= 0:
            return []
        if endpoint:
            step = (stop - start) / (num - 1)
        else:
            step = (stop - start) / num
        return [start + i * step for i in range(num)]

    def arange(start, stop=None, step=1.0) -> list:
        """
        Return evenly spaced values within a given interval. Equivalent to numpy's arange function.

        Parameters:
            start: Start of interval. The interval includes this value. The default start value is 0.
            stop: End of interval. The interval does not include this value, except in some cases where step is not an integer and floating point round-off affects the length of out.
            step: Spacing between values. For any output out, this is the distance between two adjacent values, `out[i+1] - out[i]`. The default step size is 1. If step is specified as a position argument, start must also be given.

        Returns:
            List of integer or real numbers.
        """
        if stop is None:  # If only one argument is provided, treat it as stop
            stop = start
            start = 0.0
        if step == 0:
            raise ValueError("step argument must not be zero")
        result = []
        i = start
        if step > 0:
            while i < stop:
                result.append(round(i, 10))  # Rounding to avoid floating-point precision issues
                i += step
        else:
            while i > stop:
                result.append(round(i, 10))
                i += step
        return result

    def random_point(a, b, stride=0, dtype=int):
        points = None
        if a == b:
            return a
        if stride > (b - a):
            raise ValueError("Stride is larger than the interval length.")
        elif stride == 0:
            points = Helper.arange(a, b + 1)
        else:
            points = Helper.arange(a, b + stride, stride)
        choice = random.choice(points)
        return dtype(choice)

    def range(start, end, step=1):
        if start == end:
            return [start]
        else:
            return Helper.arange(start, end + step, step)

class ErrorHandling():
    def __init__(self, max_fails:int = 10, look_back:int = 30, database:Database = None):
        self.successive_fails = 0
        self.fail_gate_open = False
        self.fail_gate_close = 0
        self.max_fails = max_fails
        self.look_back = look_back
        if self.max_fails > self.look_back:
            self.look_back = self.max_fails
        self.database = database

    def exit_action(self):
        # can be overwritten to suit specific needs
        print("[-] Successive error occurred. Exiting.")
        sys.exit(-1)

    def check(self, experiment_id:int, response:bytes, expected=b'expected', user_action=None):
        # exit if too many successive fails (including a supposedly successful memory read)
        # open fail gate, if error occurred and everything was ok previously
        if expected not in response and not self.fail_gate_open:
            self.fail_gate_open = True
            self.fail_gate_close = experiment_id + self.look_back
            self.successive_fails = 0

        # if fail gate open and error occurred, increase the fail count
        if expected not in response and self.fail_gate_open:
            self.successive_fails += 1

        # close fail gate after max_fails more experiments and check result
        if  experiment_id >= self.fail_gate_close and self.fail_gate_open:
            self.fail_gate_open = False
            if self.successive_fails >= self.max_fails:
                # delete the erroneous data points, but not the first
                if self.database is not None:
                    for eid in range(experiment_id - self.max_fails + 1, experiment_id + 1):
                        #print(f"Deleting {eid}")
                        self.database.remove_rel(eid)

                # get parameters of first erroneous experiment and store in database with extra classification
                parameters = self.database.get_parameters_of_experiment(experiment_id - self.look_back)
                response = b'error: successive error occurred'
                if self.database is not None:
                    try:
                        parameters = (experiment_id - self.look_back + 1, ) + parameters[1:-2] + ('O', response)
                        self.database.insert(*parameters)
                    except Exception as _:
                        pass
                
                # execute user action
                if user_action is None:
                    self.exit_action()
                else:
                    user_action()
                self.successive_fails = 0

            return response


from .findus import Database
from .findus import Serial
from .findus import MicroPythonScript
from .findus import PicoGlitcherInterface
from .findus import Glitcher
from .findus import PicoGlitcher
from .findus import Helper
from .findus import ErrorHandling

from .AnalogPlot import AnalogPlot
from .DebugInterface import DebugInterface
from .GeneticAlgorithm import OptimizationController
from .GlitchState import GlitchState
from .InteractivePchipEditor import InteractivePchipEditor
from .STM32Bootloader import STM32Bootloader
from .STM8Programmer import STM8Programmer

__all__ = ["Database", "Serial", "MicroPythonScript", "PicoGlitcherInterface", "Glitcher", "PicoGlitcher", "Helper", "ErrorHandling", "AnalogPlot", "DebugInterface", "OptimizationController", "GlitchState", "InteractivePchipEditor", "STM32Bootloader", "STM8Programmer"]


import time

def timed_function(f, *args, **kwargs):
    myname = str(f).split(' ')[1]
    def new_func(*args, **kwargs):
        t = time.ticks_us()
        result = f(*args, **kwargs)
        delta = time.ticks_diff(time.ticks_us(), t)
        print('Function {} Time = {:6.3f}ms'.format(myname, delta/1000))
        return result
    return new_func

class micropython:
    def native(f, *args, **kwargs):
        def new_func(*args, **kwargs):
            return f(*args, **kwargs)
        return new_func


#!/usr/bin/env python3
# Copyright (C) 2024 Dr. Matthias Kesenheimer - All Rights Reserved.
# You may use, distribute and modify this code under the terms of the GPL3 license.
#
# You should have received a copy of the GPL3 license with this file.
# If not, please write to: m.kesenheimer@gmx.net.

import time
import sys
try:
    from rd6006 import RD6006
except Exception as _:
    print("[-] Error: Library RD6006 not installed. Functions to control the external power supply not available.")
    print("    Install the rd6006 package with 'pip install rd6006'")
    sys.exit(1)

class ExternalPowerSupply:
    """
    Wrapper class for the RD6006 voltage supply to align function names with the rest of the library.
    Example usage:
    
        from findus import ExternalPowerSupply
        power_supply = ExternalPowerSupply(port="/dev/ttyACM0")
        power_supply.set_voltage(ext_power_voltage)
        print(power_supply.status())
        power_supply.power_cycle_target(power_cycle_time)

    Methods:
        __init__: Default constructor.
        status: Get the status of the voltage supply.
        set_voltage: Set the voltage.
        enable_vtarget: Enable voltage output.
        disable_vtarget: Disable voltage output
        power_cycle_target: Power cycle the target (disables output, waits a certain time, enables voltage output again).
    """

    def __init__(self, port:str):
        """
        Default constructor.
        
        Parameters:
            port: Port identifier of the voltage supply.
        """
        self.port = port
        self.r = RD6006(self.port)

    def status(self) -> str:
        """
        Get the status of the voltage supply.

        Returns:
            Status message of the voltage supply.
        """
        return self.r.status()

    def set_voltage(self, voltage:float):
        """
        Set the voltage of the power supply.
        
        Parameters:
            voltage: Desired output voltage.
        """
        try:
            #self.r.enable = False
            self.r.voltage = voltage
            #self.r.enable = True
        except Exception as _:
            pass

    def enable_vtarget(self):
        """
        Enable voltage output.
        """
        try:
            self.r.enable = True
        except Exception as _:
            pass

    def disable_vtarget(self):
        """
        Disable voltage output.
        """
        self.r.enable = False

    def power_cycle_target(self, power_cycle_time:float = 0.2):
        """
        Power cycle the target (disables output, waits a certain time, enables voltage output again).
        
        Parameters:
            power_cycle_time: Time to wait between disabling and enabling the voltage output again.
        """
        self.r.enable = False
        time.sleep(power_cycle_time)
        self.r.enable = True



#!/usr/bin/env python3
# Copyright (C) 2024 Dr. Matthias Kesenheimer - All Rights Reserved.
# You may use, distribute and modify this code under the terms of the GPL3 license.
#
# You should have received a copy of the GPL3 license with this file.
# If not, please write to: m.kesenheimer@gmx.net.

import argparse
import sys
from findus import PicoGlitcher

class Calibration():
    def __init__(self, args):
        self.args = args
        self.glitcher = PicoGlitcher()
        self.glitcher.init(port=args.rpico)
        
    def run(self):
        print("[+] Connect TRIGGER and RESET.")
        print("[+] Turn gain potentiometer of Pulse Shaping Expansion board all the way to the left (lowest gain).")
        print("[+] Connect your oscilloscope to pulse output of the Pulse Shaping Expansion board and to TRIGGER for triggering.")
        print("[+] Observe pulse.")
        print("[+] Measure maximum and minimum voltage of the generated pulse (vhigh and vlow). Take note and abort execution.")
        print("[+] Execute this script again with the found values for vhigh and vlow and check if the pulse aligns with 0V.")
        print("[+] Note that the found calibration values are stored on the Pico Glitcher persistently if this script is called with values for vlow and vhigh.")

        if self.args.vhigh != 1.0 and self.args.vlow != 0.0:
            self.glitcher.apply_calibration(self.args.vhigh, self.args.vlow, store=True)
        else:
            self.glitcher.apply_calibration(self.args.vhigh, self.args.vlow, store=False)

        while True:
            self.glitcher.do_calibration(self.args.vhigh)

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--rpico", required=True, help="rpico port", default="/dev/ttyACM0")
    parser.add_argument("--vhigh", required=False, help="The measured maximum voltage of the pulse.", type=float, default=1.0)
    parser.add_argument("--vlow", required=False, help="The measured minimum voltage of the pulse.", type=float, default=0.0)
    args = parser.parse_args()

    cal = Calibration(args)

    try:
        cal.run()
    except KeyboardInterrupt:
        if args.vhigh != 1.0 and args.vlow != 0.0:
            print("[+] Calibration saved.")
        print("\nExitting...")
        sys.exit(1)

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# Copyright (C) 2025 Dr. Matthias Kesenheimer - All Rights Reserved.
# You may use, distribute and modify this code under the terms of the GPL3 license.
#
# You should have received a copy of the GPL3 license with this file.
# If not, please write to: m.kesenheimer@gmx.net.

# programming
# > openocd -f interface/stlink.cfg -c "transport select hla_swd" -f target/stm32l0.cfg -c "init; halt; stm32l0x unlock 0; exit"
# > openocd -f interface/stlink.cfg -c "transport select hla_swd" -f target/stm32l0.cfg -c "init; halt; program toggle-led-STM32L0.elf; reset run; exit;"
# > openocd -f interface/stlink.cfg -c "transport select hla_swd" -f target/stm32l0.cfg -c "init; halt; stm32l0x lock 0; exit"
# -> power cycle the target!

# reading
# >  openocd -f interface/stlink.cfg -c "transport select hla_swd" -f target/stm32l0.cfg -c "init; dump_image dump.bin 0x08000000 0x400; exit"

# debugging (install arm-none-eabi-gdb!)
# > openocd -f interface/stlink.cfg -c "transport select hla_swd" -f target/stm32l0.cfg -c "init"
# > arm-none-eabi-gdb read-out-protection-test-STM32L0.elf
# (gdb) target extended-remote localhost:3333
# (gdb) x 0x08000000
## or
# > telnet localhost 4444
# > mdw 0x08000000

# load image to ram
# > openocd -f interface/stlink.cfg -c "transport select hla_swd" -f target/stm32l0.cfg -c "init; halt; load_image rdp-downgrade-STM32L0.elf" -c "reg sp 0x20000000" -c "reg pc 0x20000004" -c "resume; exit"

# load image to ram and execute
# > openocd -f interface/stlink.cfg -c "transport select hla_swd" -f target/stm32l0.cfg -c "init; halt; load_image rdp-downgrade-STM32L0.elf" -c "reg sp 0x20002000" -c "reg pc 0x20000fa4" -c "resume" -c "exit"

# Alternatively load with gdb:
# > openocd -f interface/stlink.cfg -c "transport select hla_swd" -f target/stm32l0.cfg -c "init; halt"
# > arm-none-eabi-gdb rdp-downgrade-STM32L0.elf
# > target remote :3333
# > load rdp-downgrade-STM32L0.elf
## Start address 0x20000fa4, load size 4764
# > x $pc
## 0x20000fa4 <Reset_Handler>:	0x4685480d
# > b *0x20000fa4
# > continue
# > x $sp
## 0x20001ff8

# read RDP level
# > openocd -f interface/stlink.cfg -c "transport select hla_swd" -f target/stm32l0.cfg -c "init" -c "mdw 0x4002201c" -c "exit"


import time
import subprocess
import re
import socket

class DebugInterface():
    def __init__(self, interface:str = "stlink", interface_config:str=None, target:str = "stm32l0", target_config:str=None, transport:str = "hla_swd", gdb_exec:str = "arm-none-eabi-gdb"):
        self.openocd_process = None
        self.gdb_process = None
        self.target_name = target
        self.socket = None
        self.gdb_exec = gdb_exec
        self.transport = transport
        if interface_config is None:
            self.interface_config = f'interface/{interface}.cfg'
        else:
            self.interface_config = interface_config
        if target_config is None:
            self.target_config = f'target/{self.target_name}.cfg'
        else:
            self.target_config = target_config

    def program_target(self, glitcher, elf_image:str = "program.elf", unlock:bool = True, rdp_level:int = 0, power_cycle_time:float = 0.1, verbose:bool = False):
        """
        TODO
        """
        if unlock:
            glitcher.reset(0.01)
            time.sleep(0.005)
            self.unlock_target(verbose)
            glitcher.power_cycle_reset(power_cycle_time)
            time.sleep(power_cycle_time)
        self.write_image(elf_image=elf_image)
        if rdp_level == 1:
            self.lock_target(verbose)
            # changes in the RDP level become active after a power-cycle
            glitcher.power_cycle_reset(power_cycle_time)
            time.sleep(power_cycle_time)

    def unlock_target(self, verbose:bool = False):
        """
        Unlock the target and remove any read-out protection.
        Attention: This will erase the targets flash!
        """
        # trunk-ignore(bandit/B607)
        # trunk-ignore(bandit/B603)
        result = subprocess.run([
            'openocd',
            '-f', self.interface_config,
            '-c', f'transport select {self.transport}',
            '-f', self.target_config,
            '-c', f'init; halt; {self.target_name}x unlock 0; exit'
            ], text=True, capture_output=True)
        if verbose:
            print(result.stdout + result.stderr)

    def lock_target(self, verbose:bool = False):
        """
        TODO
        """
        # trunk-ignore(bandit/B607)
        # trunk-ignore(bandit/B603)
        result = subprocess.run([
            'openocd',
            '-f', self.interface_config,
            '-c', f'transport select {self.transport}',
            '-f', self.target_config,
            '-c', f'init; halt; {self.target_name}x lock 0; exit'
            ], text=True, capture_output=True)
        if verbose:
            print(result.stdout + result.stderr)

    def write_image(self, elf_image:str = "program.elf", verbose:bool = False):
        """
        Write image to flash.
        """
        # trunk-ignore(bandit/B607)
        # trunk-ignore(bandit/B603)
        result = subprocess.run([
            'openocd',
            '-f', self.interface_config,
            '-c', f'transport select {self.transport}',
            '-f', self.target_config,
            '-c', f'init; halt; program {elf_image}; exit'
            ], text=True, capture_output=True)
        if verbose:
            print(result.stdout + result.stderr)

    def load_exec(self, elf_image:str = "program.elf", sp:int = 0x20002000, pc:int = 0x20000fa4, verbose:bool = False):
        """
        Load image to RAM and execute.
        Attention: Program must also be compiled to be compatible to run in RAM.
        """
        # openocd -f interface/stlink.cfg -c "transport select hla_swd" -f target/stm32l0.cfg -c "init; halt; load_image rdp-downgrade-STM32L0.elf" -c "reg sp 0x20002000" -c "reg pc 0x20000fa4" -c "resume" -c "exit"
        # trunk-ignore(bandit/B607)
        # trunk-ignore(bandit/B603)
        result = subprocess.run([
            'openocd',
            '-f', self.interface_config,
            '-c', f'transport select {self.transport}',
            '-f', self.target_config,
            '-c', f'init; halt; load_image {elf_image}',
            '-c', f'reg sp {hex(sp)}',
            '-c', f'reg pc {hex(pc)}',
            '-c', 'resume',
            '-c', 'exit',
            ], text=True, capture_output=True)
        if verbose:
            print(result.stdout + result.stderr)

    def read_image(self, bin_image:str = "memory_dump.bin", start_addr:int = 0x08000000, length:int = 0x400, verbose:bool = False):
        # trunk-ignore(bandit/B607)
        # trunk-ignore(bandit/B603)
        result = subprocess.run([
            'openocd',
            '-f', self.interface_config,
            '-c', f'transport select {self.transport}',
            '-f', self.target_config,
            '-c', f'init; dump_image {bin_image} {hex(start_addr)} {hex(length)}; exit'
            ], text=True, capture_output=True)
        if verbose:
            print(result.stdout + result.stderr)

    def test_connection(self):
        # trunk-ignore(bandit/B607)
        # trunk-ignore(bandit/B603)
        result = subprocess.run([
            'openocd',
            '-f', self.interface_config,
            '-c', f'transport select {self.transport}',
            '-f', self.target_config,
            '-c', 'init; reset run; exit'
            ], text=True, capture_output=True)
        print(result.stdout + result.stderr)

    def read_address(self, address):
        # trunk-ignore(bandit/B607)
        # trunk-ignore(bandit/B603)
        result = subprocess.run([
            'openocd',
            '-f', self.interface_config,
            '-c', f'transport select {self.transport}',
            '-f', self.target_config,
            '-c', 'init',
            '-c', f'mdw {hex(address)}',
            '-c', 'exit'
            ], text=True, capture_output=True)
        response = result.stdout + result.stderr
        return self.extract_memory_content(response=response, address=address), response

    def read_option_bytes(self):
        mem, response = self.read_address(0x4002201c)
        if mem is not None:
            return mem, response
        return 0x00, response

    def read_pcrop(self):
        optbytes, _ = self.read_option_bytes()
        pcrop = (optbytes & 0x100) >> 8
        return pcrop

    def read_rdp(self):
        optbytes, _ = self.read_option_bytes()
        rdp = (optbytes & 0xff)
        return rdp

    def read_rdp_and_pgrop(self, verbose:bool = False):
        optbytes, response = self.read_option_bytes()
        if verbose:
            print(response)
        pcrop = (optbytes & 0x100) >> 8
        rdp = (optbytes & 0xff)
        return rdp, pcrop

    def kill_process(self, port:int = 3333, verbose=False):
        # Find process using the port
        cmd = ["lsof", "-i", f":{port}"]
        #pattern = r"\b(\d+)\b"  # Regex for PID
        pattern = r"openocd\s+(\d+)"  # Regex for PID
        # Run command and search for PID
        # trunk-ignore(bandit/B603)
        result = subprocess.run(cmd, capture_output=True, text=True)
        if verbose:
            print(result)
        matches = re.findall(pattern, result.stdout)
        if matches:
            pid = matches[-1]  # Get the last match (PID)
            print(f"[+] Found process {pid} using port {port}")
            # Kill the process
            try:
                # trunk-ignore(bandit/B607)
                # trunk-ignore(bandit/B603)
                subprocess.run(["kill", "-9", pid], check=True)
                print(f"[+] Process {pid} killed successfully.")
            except Exception as e:
                print(f"[-] Error killing process: {e}")

    def attach(self, delay=0.1):
        # check if there is a dangling process that would interfere with openeocd
        self.kill_process(3333)
        # trunk-ignore(bandit/B607)
        # trunk-ignore(bandit/B603)
        self.openocd_process = subprocess.Popen([
            'openocd',
            '-f', self.interface_config,
            '-c', f'transport select {self.transport}',
            '-f', self.target_config,
            '-c', 'init'
            ], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, stdin=subprocess.DEVNULL, start_new_session=True)
        time.sleep(delay)

    def detach(self):
        if self.openocd_process is not None:
            self.openocd_process.terminate()
            self.openocd_process.wait()
            self.openocd_process = None
        if self.socket is not None:
            self.socket.close()
            self.socket = None
        if self.gdb_process is not None:
            self.gdb_process.terminate()
            self.gdb_process = None

    def gdb_load_exec(self, elf_image:str = "program.elf", timeout=0.3, verbose=False):
        # trunk-ignore(bandit/B603)
        self.gdb_process = subprocess.Popen([
            f'{self.gdb_exec}',
            '--interpreter=mi2',
            f'{elf_image}'
            ], stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        self.gdb_process.stdin.write("target remote localhost:3333\n")
        self.gdb_process.stdin.write(f"load {elf_image}\n")
        self.gdb_process.stdin.write("continue\n")
        self.gdb_process.stdin.write("detach\n")
        self.gdb_process.stdin.write("quit\n")
        self.gdb_process.stdin.flush()
        output, error = "", ""
        try:
            output, error = self.gdb_process.communicate(timeout=timeout)
        except Exception as e:
            #print(f"[-] Exception in DebugInterface:gdb_load_exec occured:\n{e}")
            pass
        finally:
            if verbose:
                print(output)
                print(error)
        self.gdb_process.terminate()

    def telnet_init(self):
        # generate a connection to the openocd telnet server
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.settimeout(1)
        self.socket.connect(('localhost', 4444))
        # receive start messages
        time.sleep(0.1)
        self.socket.recv(4096)

    def telnet_interact(self, command:str, wait_time:float = 0.01, verbose:bool = False):
        if self.openocd_process is None or self.socket is None:
            self.telnet_init()
        command += "\n"
        self.socket.sendall(command.encode("utf-8"))
        time.sleep(wait_time)
        response = self.socket.recv(4096)
        if verbose:
            print(response.decode("utf-8"))
        return response.decode("utf-8")

    def telnet_read_address(self, address:int):
        command = f"mdw {hex(address)}"
        response = self.telnet_interact(command)
        if "Previous state query failed, trying to reconnect" in response:
            response += self.telnet_interact(command)
        return self.extract_memory_content(response=response, address=address), response

    def telnet_read_image(self, bin_image:str = "memory_dump.bin", start_addr:int = 0x08000000, length:int = 0x400, verbose:bool = False):
        command = f"init; dump_image {bin_image} {hex(start_addr)} {hex(length)}; exit"
        response = self.telnet_interact(command)
        if verbose:
            print(response)

    def extract_memory_content(self, response:str, address:int = 0x00):
        if "Error: Failed to read memory at" not in response:
            match = re.search(fr'{hex(address)[2:]}:\s*([0-9A-Fa-f]+)', response)
            if match:
                return int(match.group(1), 16)
            else:
                return None

    def characterize(self, response:str, mem:int):
        # possibly ok
        if mem is not None:
            if mem != 0x00:
                return b'success: RDP inactive'
            else:
                return b'ok: read zero'
        # Error: no connection
        elif "Error: init mode failed (unable to connect to the target)" in response:
            return b'error: no connection'
        # Warning: Polling failed
        elif "Polling target" in response:
            return b'warning: polling failed'
        # Warning: Device lockup
        if "clearing lockup after double fault" in response:
            return b'warning: lock-up'
        # Warning: else
        elif "Warning" in response:
            return b'warning: default warning'
        # expected state
        elif "Error: Failed to read memory at" in response:
            return b'expected: RDP active'
        # no response
        return b'error: no response'

    def __del__(self):
        print("[+] Detaching debugger.")
        self.detach()


#!/usr/bin/env python3
# Copyright (C) 2024 Dr. Matthias Kesenheimer - All Rights Reserved.
# You may use, distribute and modify this code under the terms of the GPL3 license.
#
# You should have received a copy of the GPL3 license with this file.
# If not, please write to: m.kesenheimer@gmx.net.

import serial
from functools import reduce
import sys
import time

class STM32Bootloader:
    """
    Class that contains methods to communicate with STM32 processors in bootloader mode.
    Example usage:
    
        from findus.STM32Bootloader import STM32Bootloader
        bootcom = STM32Bootloader(port="/dev/ttyACM1")
        response = bootcom.init_bootloader()
        response = bootcom.setup_memread()
        response, mem = self.bootcom.read_memory(0x08000000, 0xFF)

    Methods:
        __init__: Default constructor.
        check_ack: Returns b'error: NACK' or b'ok: ACK' depending on the bootloader response.
        flush: Flush the serial buffers.
        init_get_id: Initializes the bootloader communication and asks the chip for its ID.
        init_bootloader: Initializes the bootloader communication via UART.
        setup_memread: Configures the bootloader to read memory from specific memory addresses.
        read_memory: Read the memory from a given memory address.
        dump_memory_to_file: Read the memory from a given memory range and write to file.
        __del__: Default deconstructor. Closes the serial communication to the bootloader.
    """
    NACK = b'\x1f'
    ACK  = b'\x79'
    verbose = False

    def __init__(self, port:str, serial_timeout:float=1, dump_address:int=0x08000000, dump_len:int=0x400):
        """
        Default constructor. Initializes the serial connection to the STM32 target (in 8e1 configuration), and stores parameter for the memory dump.
        
        Parameters:
            port: Port identifier of the STM32 target.
            dump_address: Memory address to start reading from.
            dump_len: Size of the memory to read from the device.
        Returns:
            return
        """
        print(f"[+] Opening serial port {port}.")
        self.timeout = serial_timeout
        self.ser = serial.Serial(port=port, baudrate=115200, timeout=serial_timeout, bytesize=8, parity="E", stopbits=1)
        # memory read settings
        self.current_dump_addr = dump_address
        self.current_dump_len = dump_len

    def check_ack(self) -> bytes:
        r"""
        Read a byte from serial and returns `b'error: NACK'` or `b'ok: ACK'` depending whether the device's bootloader responds with a NACK or a ACK over UART.

        Returns:
            Returns `b'error: NACK'` if `b'\\x1f'` was received, or `b'ok: ACK'` if `b'\\x79'` was received. If the target did not respond, a `b'error: no response'` is returned.
        """
        s = self.ser.read(1)
        if s == self.NACK:
            return b'error: NACK'
        elif s == self.ACK:
            return b'ok: ACK'
        return b'error: no response'

    def flush(self):
        """
        Flush the serial data buffers.
        """
        self.ser.reset_output_buffer()
        self.ser.reset_input_buffer()
        # read garbage and discard
        #self.ser.read(1024)

    def serial_read(self, nbytes=1) -> bytes:
        """
        Read a given number of byts from the serial console.
        """
        return self.ser.read(nbytes)

    def serial_write(self, bytes:bytes):
        """
        Write a byte sequence to the serial console.
        """
        return self.ser.write(bytes)

    def flush_v2(self, timeout:float = 0.01):
        """
        Flush serial data buffers with timeout.

        Parameters:
            timeout: Timeout after the serial connection stops listening. Can be tweaked to make sure all data is flushed.
        """
        self.ser.timeout = timeout
        self.ser.read(8192)
        self.ser.timeout = self.timeout

    # returns b'error: bootloader not available' if bootloader is unavailable
    # returns b'error: id command error' if reading from bootloader was not successful
    # returns b'ok' else
    def init_get_id(self) -> bytes:
        """
        Initializes the bootloader communication and asks the chip for its ID.
        
        Returns:
            Returns `b'error: bootloader not available'` if the bootloader is unavailable, `b'error: id command error'` if 'id command' was not successful, and `b'ok'` if successful.
        """
        # init bootloader
        self.ser.write(b'\x7f')
        if b'error' in self.check_ack():
            return b'error: bootloader not available'

        # get chip id command (x02: chip id, xfd: crc)
        self.ser.write(b'\x02\xfd')
        if b'error' in self.check_ack():
            return b'error: id command error'
        
        # read chip id
        s = self.ser.read(3)
        chipid = s[1:3]

        if self.verbose:
            print(f"[+] Chip ID: {chipid}")
        return b'ok'

    # returns b'ok: bootloader ok' if bootloader setup was successful (expected)
    # returns b'error: bootloader error' else
    def init_bootloader(self) -> bytes:
        """
        Initializes the bootloader communication via UART.
        
        Returns:
            Returns `b'ok: bootloader ok'` if bootloader setup was successful (expected), returns `b'error: bootloader error'` else.
        """
        # init bootloader
        self.ser.write(b'\x7f')
        s = self.ser.read(1)
        if s == self.ACK:
            return b'ok: bootloader ok'
        return b'error: bootloader error'

    # returns b'expected: RDP active' if RDP is active (expected)
    # returns b'success: RDP inactive' if glitch was successful
    # returns b'error: no response' else
    def setup_memread(self, read:bool = True) -> bytes:
        """
        Configures the bootloader to read memory from specific memory addresses.

        Parameters:
            read: Whether to read the response or not.

        Returns:
            Returns `b'expected: RDP active'` if RDP is active (expected), or `b'success: RDP inactive'` if glitch was successful
        """
        # read memory command (x11: read memory, xee: crc)
        self.ser.write(b'\x11\xee')
        if read:
            s = self.ser.read(1)
            if s == self.ACK:
                return b'success: RDP inactive'
            elif s == self.NACK:
                return b'expected: RDP active'
        return b'error: no response'

    # returns b'success: dump ok' if glitch and memory read was successful
    # returns b'ok: dump error' if glitch was successful, however memory read yielded eroneous results
    # must be used in combination with setup_memread
    def read_memory(self, start:int, size:int) -> [bytes, bytes]:
        """
        Read the memory from a given memory address.
        Must be used in combination with `setup_memread`.

        Parameters:
            start: Memory address to start reading from.
            size: Size of the memory to read from the device.
        Returns:
            Returns `b'success: dump ok'` if glitch and memory read was successful, or `b'ok: dump error'` if glitch was successful, however memory read yielded eroneous results.
        """
        # write memory address
        startb = start.to_bytes(4, 'big')
        crc = reduce(lambda x, y: x ^ y, startb, 0).to_bytes(1, 'big')
        self.ser.write(startb)
        self.ser.write(crc)
        self.ser.read(1)

        # write bytes to read
        sizeb = size.to_bytes(1, 'big')
        crc = reduce(lambda x, y: x ^ y, sizeb, 0xff).to_bytes(1, 'big')
        # write number of bytes to read
        self.ser.write(sizeb)
        self.ser.write(crc)
        self.ser.read(1)

        # read memory
        mem = self.ser.read(size)

        print(f"[+] Length of memory dump: {len(mem)}")
        print(f"[+] Content: {mem}")
        response = b'ok'
        if len(mem) == 255 and mem != b"\x00" * 255:
            response = b'success: dump ok'
        else:
            response = b'ok: dump error'
        return response, mem

    # returns b'expected: RDP active' if RDP is active (expected)
    # returns b'success: RDP inactive' if glitch was successful
    # returns b'ok: dump error' if glitch was successful but dumped memory was eroneous
    # returns b'success: dump ok' if glitch was successful and dumped memory was good
    def dump_memory_debug(self) -> [bytes, bytes]:
        # read memory (x11: read memory, xee: crc)
        self.ser.write(b'\x11\xee')
        s = self.ser.read(1)

        mem = b''
        if s == self.ACK:
            # write memory address
            self.ser.write(b'\x08\x00\x00\x00\x08')
            self.ser.read(1)
            # write number of bytes to read
            self.ser.write(b'\xff\x00')
            self.ser.read(1)

            # read memory
            mem = self.ser.read(255)

            if mem != b'\x1f' and mem != b'\x79' and mem != b'':
                print(f"[+] Length of memory dump: {len(mem)}")
                print(f"[+] Content: {mem}")
                time.sleep(5)
                return b'success: dump ok', mem
            else:
                time.sleep(5)
                return b'ok: dump error', mem

        if s == self.ACK:
            return b'success: RDP inactive', mem

        return b'expected: RDP active', mem

    # returns b'success: dump ok' if glitch and memory read was successful
    # returns b'ok: dump error' if glitch was successful, however memory read yielded eroneous results
    # must be used in combination with setup_memread
    def read_memory_debug(self, start:int, size:int) -> [bytes, bytes]:
        # write memory address
        self.ser.write(b'\x08\x00\x00\x00\x08')
        self.ser.read(1)
        # write number of bytes to read
        self.ser.write(b'\xff\x00')
        self.ser.read(1)

        # read memory
        mem = self.ser.read(size)

        print(f"[+] Length of memory dump: {len(mem)}")
        print(f"[+] Content: {mem}")
        response = b'ok'
        if len(mem) == 255 and mem != b"\x00" * 255:
            response = b'success: dump ok'
        else:
            response = b'ok: dump error'
        return response, mem

    # returns b'error: eroneous memory read' if memory read was eroneous
    # returns b'success: dump successful' if one dump was successful
    # returns b'success: dump finished' if entire memory was dummped
    def dump_memory_to_file(self, dump_filename:str) -> [bytes, bytes]:
        """
        Read the memory from a given memory range and write the memory dump to a file.
        
        Parameters:
            dump_filename: Filename to write the memory dump to.
        Returns:
            Returns `b'ok: dump error'` if memory read was eroneous, `b'success: dump successful'` if one dump was successful, and `b'success: dump finished'` if entire memory was dummped.
        """
        # read memory if RDP is inactive
        len_to_dump = 0xFF if (self.current_dump_len // 0xFF) else self.current_dump_len % 0xFF
        response, mem = self.read_memory(self.current_dump_addr, len_to_dump)

        if b'error' in response or response == b'ok: dump error':
            return b'error: eroneous memory read', mem

        # write memory dump to file
        with open(dump_filename, 'ab+') as f:
            f.write(mem)
        self.current_dump_len -= len(mem)
        print(f"[+] Dumped 0x{len(mem):x} bytes from addr 0x{self.current_dump_addr:x}, {self.current_dump_len:x} bytes left")
        self.current_dump_addr += len(mem)

        if self.current_dump_len <= 0:
            print("[+] Dump finished.")
            return b'success: dump finished', mem
        return b'success: dump successful', mem

    def __del__(self):
        """
        Default deconstructor. Closes the serial communication to the bootloader.
        """
        print("[+] Closing serial port.")
        self.ser.close()

def main(argv=sys.argv):
    com = STM32Bootloader(port=sys.argv[1])
    ret = com.init_get_id()
    print(ret)
    if ret != 0:
        sys.exit(1)

    ret = com.setup_memread()
    print(ret)
    if ret != 0:
        sys.exit(1)

    start = 0x08000000
    size  = 0xff
    ret, mem = com.read_memory(start, size)
    print(ret)
    if ret == 0:
        print(mem)
    sys.exit(1)

if __name__ == "__main__":
    main()


import time
from machine import Pin, SPI

# pin assignment
PS_TRIGGER = 5
PS_RESET = 6
PS_SPI_MISO = 16
PS_SPI_CS = 17
PS_SPI_SCK = 18
PS_SPI_MOSI = 19

# register addresses
REG_SPI_CONFIG     = 0x0000
REG_POWER_CONFIG   = 0x0001
REG_CLOCK_CONFIG   = 0x0002
REG_REF_ADJ        = 0x0003
REG_DAC_AGAIN      = 0x0007
REG_DACX_RANGE     = 0x0008
REG_DAC_RSET       = 0x000C
REG_CAL_CONFIG     = 0x000D
REG_COMP_OFFSET    = 0x000E
REG_RAM_UPDATE     = 0x001D
REG_PAT_STATUS     = 0x001E
REG_PAT_TYPE       = 0x001F
REG_PATTERN_DLY    = 0x0020
REG_DAC_DOF        = 0x0025
REG_WAV_CONFIG     = 0x0027
REG_PAT_TIMEBASE   = 0x0028
REG_PAT_PERIOD     = 0x0029
REG_DAC_PAT        = 0x002B
REG_DOUT_START_DLY = 0x002C
REG_DOUT_CONFIG    = 0x002D
REG_DAC_CST        = 0x0031
REG_DAC_DGAIN      = 0x0035
REG_SAW_CONFIG     = 0x0037
REG_DDS_TW32       = 0x003E
REG_DDS_TW1        = 0x003F
REG_DDS_PW         = 0x0043
REG_TRIG_TW_SEL    = 0x0044
REG_DDSX_CONFIG    = 0x0045
REG_TW_RAM_CONFIG  = 0x0047
REG_START_DLY      = 0x005C
REG_START_ADDR     = 0x005D
REG_STOP_ADDR      = 0x005E
REG_DDS_CYC        = 0x005F
REG_CFG_ERROR      = 0x0060
SRAM_ADDRESS_MIN   = 0x6000
SRAM_ADDRESS_MAX   = 0x6FFF

# ram update and pat status register macros
UPDATE_SETTINGS    = 0x01
MEM_ACCESS_ENABLE  = 0x04
MEM_ACCESS_DISABLE = 0x00
ENABLE_PATTERN     = 0x02
START_PATTERN      = 0x01
STOP_PATTERN       = 0x00
BUF_READ           = 0x08

# pattern control
PATTERN_RPT_CONTINOUS = 0x00
PATTERN_RPT_FINITE    = 0x01

# SPI read/write setting
SPI_READ_MASK      = 0x80
SPI_WRITE_MASK     = 0x7F

# frequency calculation constants
MASTER_CLOCK       = 125000000
FREQ_RESOLUTION    = 0x1000000

# gain calculation macros
GAIN_MAX           = 2.0
GAIN_MIN           = -2.0
GAIN_RESOLUTION    = 1024
OFFSET_RESOLUTION = 4096
DEFAULT_GAIN       = 0.5
DEFAULT_FREQUENCY  = 100000

# wave output selection macros
WAVE_SINE              = 0x00
WAVE_COSINE            = 0x01
WAVE_TRIANGLE          = 0x02
WAVE_POSITIVE_SAWTOOTH = 0x03
WAVE_NEGATIVE_SAWTOOTH = 0x04

# wave config macros
WAV_CFG_PRESTORE_CST         = 0x00
WAV_CFG_PRESTORE_SAWTOOTH    = 0x10
WAV_CFG_PRESTORE_PSEUDO      = 0x20
WAV_CFG_PRESTORE_DDS         = 0x30
WAV_CFG_WAVE_FROM_RAM        = 0x00
WAV_CFG_WAVE_PRESTORED       = 0x01
WAV_CFG_WAVE_PRESTORED_DELAY = 0x02
WAV_CFG_WAVE_PRESTORED_RAM   = 0x03
DDSX_CFG_ENABLE_COSINE       = 0x08
SAW_CFG_RAMP_UP              = 0x00
SAW_CFG_RAMP_DOWN            = 0x01
SAW_CFG_TRIANGLE             = 0x02
SAW_CFG_NO_WAVE              = 0x03
SAW_CFG_STEP_1               = 0x04

def usleep(x:int):
    time.sleep(x/1000000.0)

def bit_not(n, numbits=8):
    return (1 << numbits) - 1 - n

class AD910X():
    def __init__(self):
        self.pin_cs = Pin(PS_SPI_CS, mode=Pin.OUT, value=1, pull=Pin.PULL_UP)
        self.pin_reset = Pin(PS_RESET, mode=Pin.OUT, value=1, pull=Pin.PULL_UP)
        self.pin_trigger = Pin(PS_TRIGGER, mode=Pin.OUT, value=1, pull=Pin.PULL_UP)
        self.spi = SPI(0,
                  baudrate=1000000, # 100000 oder 1000000
                  polarity=0,
                  phase=0,
                  bits=8, # 8 oder 16
                  firstbit=SPI.MSB,
                  sck=Pin(PS_SPI_SCK),
                  mosi=Pin(PS_SPI_MOSI),
                  miso=Pin(PS_SPI_MISO))

    def init(self):
        #self.spi_read_register(REG_SPI_CONFIG)
        # TODO: do a proper calibration!
        #self.spi_write_register(REG_DAC_AGAIN, 0x4000)
        #self.spi_write_register(REG_DAC_RSET, 0x1F00) # DAC_RSET_CAL = 11111
        #self.spi_write_register(REG_DOUT_START_DLY, 0x0003)
        #self.spi_write_register(REG_COMP_OFFSET, ) offset
        pass

    def get_trigger_pin(self):
        """
        Get the trigger pin.
        """
        return self.pin_trigger

    def init_trigger_pin(self):
        """
        Re-initialize the trigger pin. This is required if you want to regain control of the pin outside the state machine.
        """
        self.pin_trigger = Pin(PS_TRIGGER, mode=Pin.OUT, value=1, pull=Pin.PULL_UP)
        return self.pin_trigger

    def spi_write_registers(self, addr:int, data:list[int]):
        """
        Write a list of 16-bit data to AD910x SPI/SRAM register

        Parameters:
            addr: 16-bit SPI/SRAM start address
            data: list of 16-bit data to be written
        """
        tx_buf = [0] * (2 * len(data) + 2)
        tx_buf[0] = ((((addr >> 8) & 0xFF) & SPI_WRITE_MASK) & 0xFF)
        tx_buf[1] = (addr & 0xFF)
        for cnt in range(len(data)):
            tx_buf[( cnt * 2 ) + 2] = ((data[cnt] >> 8) & 0xFF)
            tx_buf[( cnt * 2 ) + 3] = (data[cnt] & 0xFF)
        self.pin_cs.value(0)
        self.spi.write(bytes(tx_buf))
        self.pin_cs.value(1)
        #usleep(1)

    def spi_read_registers(self, addr:int, length:int) -> list[int]:
        """
        Read a list of 16-bit data from AD910x SPI/SRAM register.

        Parameters:
            addr: 16-bit SPI/SRAM start address.
            length: number of registers to read.

        Returns:
            16-bit data returned by AD910x
        """
        tx_buf = [0] * 2
        tx_buf[0] = (((( addr >> 8 ) & 0xFF ) | SPI_READ_MASK) & 0xFF)
        tx_buf[1] = (addr & 0xFF )
        self.pin_cs.value(0)
        self.spi.write(bytes(tx_buf))
        rx_buf = list(self.spi.read(length * 2))
        self.pin_cs.value(1)
        if len(rx_buf) == 0:
            return []
        data_out = [0] * length
        for cnt in range(0, length):
            data_out[cnt] = (rx_buf[cnt * 2] << 8) | rx_buf[(cnt * 2) + 1]
        #usleep(1)
        return data_out

    def spi_write_register_new(self, addr:int, data:int):
        """
        Write 16-bit data to AD910x SPI/SRAM register

        Parameters:
            addr: 16-bit SPI/SRAM address
            data: 16-bit data to be written to register address
        """
        self.pin_cs.value(0)
        addr_bytes = addr.to_bytes(2, 'big')
        self.spi.write(addr_bytes)
        data_bytes = data.to_bytes(2, 'big')
        self.spi.write(data_bytes)
        self.pin_cs.value(1)
        usleep(1)

    def spi_read_register_new(self, addr:int) -> int:
        """
        Read 16-bit data from AD910x SPI/SRAM register.

        Parameters:
            addr: 16-bit SPI/SRAM address.

        Returns:
            16-bit data returned by AD910x.
        """
        read_addr = 0x8000 + addr
        self.pin_cs.value(0)
        self.spi.write(read_addr.to_bytes(2, 'big'))
        data = self.spi.read(1)
        self.pin_cs.value(1)
        usleep(1)
        return int.from_bytes(data, 'big')

    def spi_write_register(self, addr:int, data:int):
        """
        Write 16-bit data to AD910x SPI/SRAM register

        Parameters:
            addr: 16-bit SPI/SRAM address
            data: 16-bit data to be written to register address
        """
        self.spi_write_registers(addr, [data])

    def spi_read_register(self, addr:int) -> int:
        """
        Read 16-bit data from AD910x SPI/SRAM register.

        Parameters:
            addr: 16-bit SPI/SRAM address.

        Returns:
            16-bit data returned by AD910x.
        """
        data = self.spi_read_registers(addr, 1)
        if len(data) > 0:
            return data[0]
        return 0x00

    def reset(self):
        """
        Reset AD910x SPI registers to default values
        """
        self.pin_reset.low()
        usleep(100)
        self.pin_reset.high()
        usleep(100000)

    def trigger_high(self):
        """
        Set trigger pin high (disables pattern generation).
        """
        self.pin_trigger.high()

    def trigger_low(self):
        """
        Set trigger pin low (enables pattern generation).
        """
        self.pin_trigger.low()

    def print_data(self, addr:int, data:int, addrname:str = ""):
        """
        Print register address and data in hexadecimal format

        Parameters:
            addr: 16-bit SPI/SRAM register address
            data: 16-bit data
        """
        if addrname == "":
            print(f'0x{addr:0>4X}: 0x{data:0>4X} = 0b{data:0>16b}')
        else:
            print(f'0x{addr:0>4X} ({addrname}):\t0x{data:0>4X} = 0b{data:0>16b}')

    def write_sram(self, addr:int, data:list[int]):
        """
        Write data to SRAM.

        Parameters:
            addr: 16-bit SPI/SRAM register start address.
            data: array of 16-bit data to be written to SRAM.
        """
        if (addr < SRAM_ADDRESS_MIN) or (addr > SRAM_ADDRESS_MAX) or ((addr + len(data)) > (SRAM_ADDRESS_MAX + 1)):
            raise Exception("SRAM address not in range [0x6000, 0x6FFF]")
        self.spi_write_register(REG_PAT_STATUS, MEM_ACCESS_ENABLE)
        for cnt in range(0, len(data)):
            if data[cnt] > 8190:
                data[cnt] = 8190
            elif data[cnt] < -8190:
                data[cnt] = -8190
            self.spi_write_register(addr + cnt, data[cnt] << 2)
        self.spi_write_register(REG_PAT_STATUS, MEM_ACCESS_DISABLE)

    def write_sram_from_start(self, data:list[int]):
        """
        Write data to SRAM from start address.

        Parameters:
            data: array of 16-bit data to be written to SRAM.
        """
        if len(data) > 4096:
            raise Exception("SRAM data too large.")
        self.write_sram(SRAM_ADDRESS_MIN, data)

    def set_const(self, value:int, length:int = 4096):
        if value > 8190:
            value = 8190
        elif value < -8190:
            value = -8190
        self.spi_write_register(REG_PAT_STATUS, MEM_ACCESS_ENABLE)
        for addr in range(SRAM_ADDRESS_MIN, SRAM_ADDRESS_MIN + length):
            self.spi_write_register(addr, value << 2)
        self.spi_write_register(REG_PAT_STATUS, MEM_ACCESS_DISABLE)

    def read_sram(self, addr:int, length:int) -> list[int]:
        """
        Read array of 16-bit data from SRAM.

        Parameters:
            length: number of SRAM addresses to be read from.

        Returns:
            Array of 16-bit data read from SRAM.
        """
        if (addr < SRAM_ADDRESS_MIN) or (addr > SRAM_ADDRESS_MAX) or ((addr + length) > (SRAM_ADDRESS_MAX + 1)):
            raise Exception("SRAM address not in range [0x6000, 0x6FFF]")
        self.spi_write_register(REG_PAT_STATUS, MEM_ACCESS_ENABLE | BUF_READ)
        data = [0] * length
        for cnt in range(0, length):
            data[cnt] = self.spi_read_register(addr + cnt) >> 2
        self.spi_write_register(REG_PAT_STATUS, MEM_ACCESS_DISABLE)
        return data

    def print_sram(self, length:int):
        """
        Read from SRAM and print data.

        Parameters:
            n: number of SRAM addresses to be read from
        """
        data = self.read_sram(SRAM_ADDRESS_MIN, length)
        for i in range(len(data)):
            self.print_data(SRAM_ADDRESS_MIN + i, data[i])

    def start_pattern(self):
        """
        Start pattern generation by setting AD910x trigger pin to 0.
        """
        self.spi_write_register(REG_PAT_STATUS, START_PATTERN)
        self.pin_trigger.low()

    def stop_pattern(self):
        """
        Stop pattern generation by setting AD910x trigger pin to 1.
        """
        self.spi_write_register(REG_PAT_STATUS, STOP_PATTERN)
        self.pin_trigger.high()

    def update_settings(self):
        """
        Update the settings of AD910x.
        """
        self.spi_write_register(REG_RAM_UPDATE, UPDATE_SETTINGS)

    def set_frequency(self, freq:int):
        """
        Set the frequency of the waveform.

        Parameters:
            freq: Frequency to be set.
        """
        if freq > MASTER_CLOCK:
            raise Exception("Frequency not supported.")
        freq_tmp = int((freq / (MASTER_CLOCK / FREQ_RESOLUTION))) & 0xFFFFFFFF
        tw_msb = ((freq_tmp >> 8) & 0xFFFF)
        tw_lsb = ((freq_tmp << 8 ) & 0xFF00)
        self.spi_write_register(REG_DDS_TW32, tw_msb)
        self.spi_write_register(REG_DDS_TW1, tw_lsb)

    def set_gain(self, gain:float):
        """
        Set the gain of the DAC.

        Parameters:
            gain: Gain to be set.
        """
        if gain > GAIN_MAX or gain < GAIN_MIN:
            raise Exception("Gain value not supported.")
        gain_tmp = (int((gain * GAIN_RESOLUTION)) << 4) & 0xFFFF
        self.spi_write_register(REG_DAC_DGAIN, gain_tmp)

    def set_offset(self, offset:int):
        """
        Set the offset of the DAC.

        Parameters:
            offset: Offset to be set.
        """
        if offset < 0 or offset > 4096:
            raise Exception("Offset value not supported.")
        self.spi_write_register(REG_DAC_DOF, offset)

    def update_sram(self, pulse_number_of_points:int):
        self.spi_write_register(REG_START_ADDR, 0x0000) # start SRAM addr to read data from
        stop_addr = (((pulse_number_of_points & 0x0FFF) - 1) << 4) & 0xFFF0
        self.spi_write_register(REG_STOP_ADDR, stop_addr)  # stop SRAM addr
        self.spi_write_register(REG_RAM_UPDATE, UPDATE_SETTINGS)
        self.spi_write_register(REG_PAT_STATUS, START_PATTERN)

    def set_pulse_output_oneshot(self):
        """
        Configure the DDS to output one defined pulse.
        """
        # update settings
        self.spi_write_register(REG_PAT_TYPE, PATTERN_RPT_FINITE) # pattern is emitted a finite amount of times
        self.spi_write_register(REG_DAC_PAT, 0x0001) # repeat pattern once
        self.spi_write_register(REG_WAV_CONFIG, WAV_CFG_PRESTORE_DDS) # output from DDS
        self.spi_write_register(REG_PAT_TIMEBASE, 0x0111) # HOLD = 1, PAT_PERIOD_BASE = 1, START_DELAY_BASE = 1; TODO: set the time base, TODO: HOLD = 0 for faster sampling?
        #self.spi_write_register(REG_PATTERN_DLY, 0x000E) # TODO: control this by the delay parameter
        #self.spi_write_register(REG_START_DLY, 0x0003) # TODO: OR: control this by the delay parameter

    def set_pulse_output_continous(self):
        """
        Configure the DDS to output a defined pulse continously.
        """
        # update settings
        self.spi_write_register(REG_PAT_TYPE, PATTERN_RPT_CONTINOUS)
        self.spi_write_register(REG_DAC_PAT, 0x0000)
        self.spi_write_register(REG_WAV_CONFIG, WAV_CFG_PRESTORE_DDS) # output from DDS
        self.spi_write_register(REG_PAT_TIMEBASE, 0x0111)

    def set_wave_output(self, wave:int):
        """
        Set the type of wave to output.

        Parameters:
            wave: Waveform identifier.
        """
        if wave < WAVE_SINE or wave > WAVE_NEGATIVE_SAWTOOTH:
            raise Exception("Waveform identifier not supported.")
        reg_address = REG_WAV_CONFIG
        self.stop_pattern()
        reg_data = self.spi_read_register(reg_address) & 0xFF00

        if wave < WAVE_TRIANGLE:
            reg_data |= (WAV_CFG_PRESTORE_DDS | WAV_CFG_WAVE_PRESTORED)
            self.spi_write_register(reg_address, reg_data)
            reg_data = self.spi_read_register(REG_DDSX_CONFIG)
            if wave == WAVE_COSINE:
                reg_data |= DDSX_CFG_ENABLE_COSINE
            else:
                reg_data &= bit_not((DDSX_CFG_ENABLE_COSINE & 0xFFFF))
            self.spi_write_register(REG_DDSX_CONFIG, reg_data)
        else:
            reg_data |= ((WAV_CFG_PRESTORE_SAWTOOTH | WAV_CFG_WAVE_PRESTORED))
            self.spi_write_register(reg_address, reg_data)
            reg_address = REG_SAW_CONFIG
            reg_data = self.spi_read_register(reg_address) & 0xFF00

            if wave == WAVE_TRIANGLE:
                reg_data |= ((SAW_CFG_TRIANGLE | SAW_CFG_STEP_1))
            elif wave == WAVE_POSITIVE_SAWTOOTH:
                reg_data |= ((SAW_CFG_RAMP_UP | SAW_CFG_STEP_1))
            else:
                reg_data |= ((SAW_CFG_RAMP_DOWN | SAW_CFG_STEP_1))
            self.spi_write_register(reg_address, reg_data)
        self.update_settings()
        self.start_pattern()
        return self.spi_read_register(REG_CFG_ERROR)

    def update_regs(self, data):
        regadd = [0x0000, 0x0001, 0x0002, 0x0003, 0x0004, 0x0005, 0x0006, 0x0007, 0x0008, 0x0009, 0x000a, 0x000b, 0x000c, 0x000d, 0x000e, 0x001f, 0x0020, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027, 0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f, 0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037, 0x003e, 0x003f, 0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0047, 0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057, 0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x005e, 0x005f, 0x001e, 0x001d]
        for i in range(len(regadd)):
            self.spi_write_register(regadd[i], data[i])

    def set_pulse_output_test(self):
        #self.stop_pattern()
        regval = [0x0000, 0x0e00, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x4000, 0x0000, 0x0000, 0x0000, 0x0000, 0x1f00, 0x0000, 0x0000, 0x0001, 0x000E, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x3030, 0x0111, 0xffff, 0x0000, 0x0101, 0x0003, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x4000, 0x0000, 0x0200, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0fa0, 0x0000, 0xfff0, 0x0100, 0x0001, 0x0001]
        self.update_regs(regval)

    def debug_registers(self):
        self.print_data(REG_WAV_CONFIG, self.spi_read_register(REG_WAV_CONFIG), "REG_WAV_CONFIG")
        self.print_data(REG_DDSX_CONFIG, self.spi_read_register(REG_DDSX_CONFIG), "REG_DDSX_CONFIG")
        self.print_data(REG_DAC_DGAIN, self.spi_read_register(REG_DAC_DGAIN), "REG_DAC_DGAIN")
        self.print_data(REG_DDS_TW32, self.spi_read_register(REG_DDS_TW32), "REG_DDS_TW32")
        self.print_data(REG_DDS_TW1, self.spi_read_register(REG_DDS_TW1), "REG_DDS_TW1")
        self.print_data(REG_DOUT_START_DLY, self.spi_read_register(REG_DOUT_START_DLY), "REG_DOUT_START_DLY")
        self.print_data(REG_PATTERN_DLY, self.spi_read_register(REG_PATTERN_DLY), "REG_PATTERN_DLY")
        self.print_data(REG_PAT_STATUS, self.spi_read_register(REG_PAT_STATUS), "REG_PAT_STATUS")
        self.print_data(REG_RAM_UPDATE, self.spi_read_register(REG_RAM_UPDATE), "REG_RAM_UPDATE")
        self.print_data(REG_CFG_ERROR, self.spi_read_register(REG_CFG_ERROR), "REG_CFG_ERROR")
        print()

    def set_wave_output_test(self):
        self.pin_trigger.high()

        gain = 0.5
        gain_tmp = (int((gain * GAIN_RESOLUTION)) << 4) & 0xFFFFFFFF
        self.spi_write_register(REG_DAC_DGAIN, gain_tmp)

        freq = 1000000
        freq_tmp = int((freq / (MASTER_CLOCK / FREQ_RESOLUTION))) & 0xFFFFFFFF
        tw_msb = ((freq_tmp >> 8) & 0xFFFF)
        tw_lsb = ((freq_tmp << 8 ) & 0xFF00)
        self.spi_write_register(REG_DDS_TW32, tw_msb)
        self.spi_write_register(REG_DDS_TW1, tw_lsb)

        reg_data = self.spi_read_register(REG_WAV_CONFIG) & 0xFF00
        reg_data |= (WAV_CFG_PRESTORE_DDS | WAV_CFG_WAVE_PRESTORED) # debug: 0x31 = 0b110001
        self.spi_write_register(REG_WAV_CONFIG, reg_data)

        reg_data = self.spi_read_register(REG_DDSX_CONFIG)
        reg_data |= DDSX_CFG_ENABLE_COSINE
        self.spi_write_register(REG_DDSX_CONFIG, reg_data) # debug: 0x08

        self.spi_write_register(REG_RAM_UPDATE, 0x0001)
        self.spi_write_register(REG_PAT_STATUS, 0x0001)
        self.debug_registers()
        self.pin_trigger.low()

        return self.spi_read_register(REG_CFG_ERROR)